import os
import sys


def strtobool(val: str) -> bool:
    val = val.lower()
    if val in ("y", "yes", "t", "true", "on", "1"):
        return True
    elif val in ("n", "no", "f", "false", "off", "0"):
        return False
    else:
        return False


def use_emojis() -> bool:
    # All emojis that pipx might possibly use
    emoji_test_str = "✨🌟⚠️😴⣷⣯⣟⡿⢿⣻⣽⣾"
    try:
        emoji_test_str.encode(sys.stderr.encoding)
        platform_emoji_support = True
    except UnicodeEncodeError:
        platform_emoji_support = False
    use_emoji = os.getenv("PIPX_USE_EMOJI")
    if use_emoji is None:
        use_emoji = str(os.getenv("USE_EMOJI", platform_emoji_support))
    return strtobool(use_emoji)


EMOJI_SUPPORT = use_emojis()

if EMOJI_SUPPORT:
    stars = "✨ 🌟 ✨"
    hazard = "⚠️"
    error = "⛔"
    sleep = "😴"
else:
    stars = ""
    hazard = ""
    error = ""
    sleep = ""



import logging
import os
import shutil
import subprocess
import sys
from pathlib import Path

from packaging import version

from pipx.constants import WINDOWS, FetchPythonOptions
from pipx.standalone_python import download_python_build_standalone
from pipx.util import PipxError

logger = logging.getLogger(__name__)


def has_venv() -> bool:
    try:
        import venv  # noqa: F401, PLC0415
    except ImportError:
        return False
    return True


class InterpreterResolutionError(PipxError):
    def __init__(self, source: str, version: str, wrap_message: bool = True):
        self.source = source
        self.version = version
        potentially_path = "/" in version
        potentially_pylauncher = "python" not in version and not potentially_path

        message = (
            f"No executable for the provided Python version '{version}' found in {source}."
            " Please make sure the provided version is "
        )
        if source == "py launcher":
            message += "listed when running `py --list`."
        if source == "PATH":
            message += "on your PATH or the file path is valid. "
            if potentially_path:
                message += "The provided version looks like a path, but no executable was found there."
            if potentially_pylauncher:
                message += (
                    "The provided version looks like a version, "
                    "but both the python command and the Python Launcher were not found on PATH."
                )
        if source == "the python-build-standalone project":
            message += "listed in https://github.com/astral-sh/python-build-standalone/releases/latest."
        if source == "PIPX_DEFAULT_PYTHON":
            message += (
                "on your PATH or the file path is valid. "
                "Try setting it to an executable name (e.g. python3.10) "
                "or a full path (e.g. /usr/bin/python3.10)."
            )
        super().__init__(message, wrap_message)


def find_unix_command_python(python_version: str) -> str | None:
    try:
        parsed_python_version = version.parse(python_version)
    except version.InvalidVersion:
        logger.info(f"Invalid Python version: {python_version}")
        return None

    if (
        parsed_python_version.epoch != 0
        or parsed_python_version.is_devrelease
        or parsed_python_version.is_postrelease
        or parsed_python_version.is_prerelease
    ):
        logger.info(f"Unsupported Python version: {python_version}")
        return None

    # Python command could be `python3` or `python3.x` without micro version component
    python_command = f"python{'.'.join(python_version.split('.')[:2])}"

    python_path = shutil.which(python_command)
    if not python_path:
        logger.info(f"Command `{python_command}` was not found on the system")
        return None

    if parsed_python_version.micro != 0:
        logger.warning(
            f"The command '{python_command}' located at '{python_path}' will be used. "
            f"It may not match the specified version {python_version} at the micro/patch level."
        )

    return python_path


def _fetch_standalone_interpreter(python_version: str) -> str:
    try:
        return download_python_build_standalone(python_version)
    except PipxError as e:
        raise InterpreterResolutionError(source="the python-build-standalone project", version=python_version) from e


def find_python_interpreter(python_version: str, fetch_python: FetchPythonOptions = FetchPythonOptions.NEVER) -> str:
    if fetch_python == FetchPythonOptions.ALWAYS:
        if os.sep in python_version or "/" in python_version or Path(python_version).is_file():
            raise PipxError(
                f"--fetch-python={FetchPythonOptions.ALWAYS} requires a Python version "
                f"(e.g. '3.13'), got '{python_version}'."
            )
        return _fetch_standalone_interpreter(python_version)

    if Path(python_version).is_file() or shutil.which(python_version):
        return python_version

    if not WINDOWS:
        python_unix_command = find_unix_command_python(python_version)
        if python_unix_command:
            return python_unix_command

    try:
        py_executable = find_py_launcher_python(python_version)
        if py_executable:
            return py_executable
    except (subprocess.CalledProcessError, FileNotFoundError) as e:
        if fetch_python != FetchPythonOptions.MISSING:
            raise InterpreterResolutionError(source="py launcher", version=python_version) from e

    if fetch_python == FetchPythonOptions.MISSING:
        return _fetch_standalone_interpreter(python_version)

    raise InterpreterResolutionError(source="PATH", version=python_version)


# The following code was copied from https://github.com/uranusjr/pipx-standalone
# which uses the same technique to build a completely standalone pipx
# distribution.
#
# If we are running under the Windows embeddable distribution,
# venv isn't available (and we probably don't want to use the
# embeddable distribution as our applications' base Python anyway)
# so we try to locate the system Python and use that instead.


def find_py_launcher_python(python_version: str | None = None) -> str | None:
    py = shutil.which("py")
    if py and python_version:
        python_semver = python_version
        if python_version.startswith("python"):
            logger.warning(
                "Removing `python` from the start of the version, as pylauncher just expects the semantic version"
            )
            python_semver = python_semver.lstrip("python")
        py = subprocess.run(
            [py, f"-{python_semver}", "-c", "import sys; print(sys.executable)"],
            capture_output=True,
            text=True,
            check=True,
        ).stdout.strip()
    return py


def _find_default_windows_python() -> str:
    if has_venv():
        return sys.executable
    python = find_py_launcher_python() or shutil.which("python")

    if python is None:
        raise PipxError("No suitable Python found")

    # If the path contains "WindowsApps", it's the store python
    if "WindowsApps" not in python:
        return python

    # Special treatment to detect Windows Store stub.
    # https://twitter.com/zooba/status/1212454929379581952

    proc = subprocess.run([python, "-V"], stdout=subprocess.PIPE, stderr=subprocess.DEVNULL, check=False)
    if proc.returncode != 0:
        # Cover the 9009 return code preemptively.
        raise PipxError("No suitable Python found")
    if not proc.stdout.strip():
        # A real Python should print version, Windows Store stub won't.
        raise PipxError("No suitable Python found")
    return python  # This executable seems to work.


def _get_sys_executable() -> str:
    if WINDOWS:
        return _find_default_windows_python()
    else:
        return str(Path(sys.executable).resolve())


def _resolve_python(python: str) -> str:
    if (path := Path(python)).is_file():
        return str(path.resolve())
    if found := shutil.which(python):
        return found
    if not WINDOWS and (found := find_unix_command_python(python)):
        return found
    raise InterpreterResolutionError(source="PIPX_DEFAULT_PYTHON", version=python)


env_default_python = os.environ.get("PIPX_DEFAULT_PYTHON")

if not env_default_python:
    DEFAULT_PYTHON = _get_sys_executable()
else:
    DEFAULT_PYTHON = _resolve_python(env_default_python)



import json
import sys
from collections.abc import Iterator
from pathlib import Path

from packaging.utils import canonicalize_name

from pipx import commands, paths
from pipx.backends import PIP
from pipx.commands.common import package_name_from_spec, run_post_install_actions
from pipx.constants import (
    EXIT_CODE_INSTALL_VENV_EXISTS,
    EXIT_CODE_OK,
    ExitCode,
)
from pipx.emojis import sleep
from pipx.interpreter import DEFAULT_PYTHON
from pipx.pipx_metadata_file import PackageInfo, PipxMetadata, load_spec_file
from pipx.util import PipxError, pipx_wrap
from pipx.venv import Venv, VenvContainer


def install(
    venv_dir: Path | None,
    package_names: list[str] | None,
    package_specs: list[str],
    local_bin_dir: Path,
    local_man_dir: Path,
    python: str | None,
    pip_args: list[str],
    venv_args: list[str],
    verbose: bool,
    *,
    force: bool,
    reinstall: bool,
    include_dependencies: bool,
    preinstall_packages: list[str] | None,
    suffix: str = "",
    python_flag_passed: bool = False,
    backend: str | None = None,
    env_backend: str | None = None,
) -> ExitCode:
    """Returns pipx exit code."""
    # package_spec is anything pip-installable, including package_name, vcs spec,
    #   zip file, or tar.gz file.

    python = python or DEFAULT_PYTHON

    package_names = package_names or []
    if len(package_names) != len(package_specs):
        package_names = [
            package_name_from_spec(
                package_spec,
                python,
                pip_args=pip_args,
                verbose=verbose,
                backend=backend,
                env_backend=env_backend,
            )
            for package_spec in package_specs
        ]

    for package_name, package_spec in zip(package_names, package_specs, strict=False):
        if venv_dir is None:
            venv_container = VenvContainer(paths.ctx.venvs)
            venv_dir = venv_container.get_venv_dir(f"{package_name}{suffix}")

        try:
            exists = venv_dir.exists() and bool(next(venv_dir.iterdir()))
        except StopIteration:
            exists = False

        # ``pipx install pip`` always uses pip (uv venvs ship no pip). Override
        # only the implicit env path; ``--backend uv`` still falls through to
        # ``assert_not_pip_under_uv`` so an explicit conflict fails loudly.
        install_backend, install_env_backend = backend, env_backend
        if canonicalize_name(package_name) == "pip":
            install_backend = backend or PIP
            install_env_backend = None

        venv = Venv(
            venv_dir,
            python=python,
            verbose=verbose,
            backend=install_backend,
            env_backend=install_env_backend,
        )
        venv.check_upgrade_shared_libs(pip_args=pip_args, verbose=verbose)
        if exists:
            if not reinstall and force and python_flag_passed:
                print(
                    pipx_wrap(
                        f"""
                        --python is ignored when --force is passed.
                        If you want to reinstall {package_name} with {python},
                        run `pipx reinstall {package_spec} --python {python}` instead.
                        """
                    )
                )
            if force:
                print(f"Installing to existing venv {venv.name!r}")
                pip_args = ["--force-reinstall"] + pip_args
            else:
                installed_version = venv.pipx_metadata.main_package.package_version
                version_info = f" ({installed_version})" if installed_version else ""
                print(
                    pipx_wrap(
                        f"""
                        {venv.name!r}{version_info} already seems to be installed. Not
                        modifying existing installation in '{venv_dir}'.
                        Pass '--force' to force installation, or use
                        'pipx upgrade {venv.name}' to upgrade.
                        """
                    )
                )
                if len(package_specs) == 1:
                    return EXIT_CODE_INSTALL_VENV_EXISTS
                # Reset venv_dir to None ready to install the next package in the list
                venv_dir = None
                continue

        try:
            override_shared = canonicalize_name(package_name) == "pip"
            venv.create_venv(venv_args, pip_args, override_shared)
            for dependency in preinstall_packages or []:
                venv.upgrade_package_no_metadata(dependency, [])
            venv.install_package(
                package_name=package_name,
                package_or_url=package_spec,
                pip_args=pip_args,
                include_dependencies=include_dependencies,
                include_apps=True,
                is_main_package=True,
                suffix=suffix,
            )
            run_post_install_actions(
                venv,
                package_name,
                local_bin_dir,
                local_man_dir,
                venv_dir,
                include_dependencies,
                force=force,
            )
        except (Exception, KeyboardInterrupt):
            print()
            venv.remove_venv()
            raise

        # Reset venv_dir to None ready to install the next package in the list
        venv_dir = None

    # Any failure to install will raise PipxError, otherwise success
    return EXIT_CODE_OK


def extract_venv_metadata(spec_metadata_file: Path) -> Iterator[PipxMetadata]:
    try:
        spec = load_spec_file(spec_metadata_file)
    except json.decoder.JSONDecodeError as exc:
        raise PipxError("The spec metadata file is an invalid JSON file.") from exc

    venvs_metadata_dict = spec.get("venvs")
    if not venvs_metadata_dict:
        raise PipxError("No packages found in the spec metadata file.")
    if not isinstance(venvs_metadata_dict, dict):
        raise PipxError("The spec metadata file is invalid.")

    for package_path_name, entry in venvs_metadata_dict.items():
        venv_dir = paths.ctx.venvs.joinpath(package_path_name)
        venv_metadata = PipxMetadata(venv_dir, read=False)
        venv_metadata.from_dict(entry["metadata"])
        yield venv_metadata


def generate_package_spec(package_info: PackageInfo) -> str:
    """Generate more precise package spec from package info."""
    if not package_info.package_or_url:
        raise PipxError(f"A package spec is not available for {package_info.package}")

    if package_info.package == package_info.package_or_url:
        return f"{package_info.package}=={package_info.package_version}"
    return package_info.package_or_url


def get_python_interpreter(
    source_interpreter: Path | None,
) -> str | None:
    """Get appropriate python interpreter."""
    if source_interpreter is not None and source_interpreter.is_file():
        return str(source_interpreter)

    print(
        pipx_wrap(
            f"""
            The exported python interpreter '{source_interpreter}' is ignored
            as not found.
            """
        )
    )

    return None


def install_all(
    spec_metadata_file: Path,
    local_bin_dir: Path,
    local_man_dir: Path,
    python: str | None,
    pip_args: list[str],
    venv_args: list[str],
    verbose: bool,
    *,
    force: bool,
    backend: str | None = None,
    env_backend: str | None = None,
) -> ExitCode:
    """Return pipx exit code."""
    venv_container = VenvContainer(paths.ctx.venvs)
    failed: list[str] = []
    installed: list[str] = []

    for venv_metadata in extract_venv_metadata(spec_metadata_file):
        # Install the main package
        main_package = venv_metadata.main_package
        venv_dir = venv_container.get_venv_dir(f"{main_package.package}{main_package.suffix}")
        try:
            install(
                venv_dir,
                None,
                [generate_package_spec(main_package)],
                local_bin_dir,
                local_man_dir,
                python or get_python_interpreter(venv_metadata.source_interpreter),
                pip_args,
                venv_args,
                verbose,
                force=force,
                reinstall=False,
                include_dependencies=main_package.include_dependencies,
                preinstall_packages=[],
                suffix=main_package.suffix,
                backend=backend or venv_metadata.backend,
                env_backend=env_backend,
            )

            # Install the injected packages
            for inject_package in venv_metadata.injected_packages.values():
                commands.inject(
                    venv_dir=venv_dir,
                    package_specs=[generate_package_spec(inject_package)],
                    requirement_files=[],
                    pip_args=pip_args,
                    verbose=verbose,
                    include_apps=inject_package.include_apps,
                    include_dependencies=inject_package.include_dependencies,
                    force=force,
                    suffix=inject_package.suffix == main_package.suffix,
                )
        except PipxError as e:
            print(e, file=sys.stderr)
            failed.append(venv_dir.name)
        else:
            installed.append(venv_dir.name)
    if len(installed) == 0:
        print(f"No packages installed after running 'pipx install-all {spec_metadata_file}' {sleep}")
    if len(failed) > 0:
        raise PipxError(f"The following package(s) failed to install: {', '.join(failed)}")
    # Any failure to install will raise PipxError, otherwise success
    return EXIT_CODE_OK


__all__ = [
    "extract_venv_metadata",
    "generate_package_spec",
    "get_python_interpreter",
    "install",
    "install_all",
]



import ast
import configparser
import json
import logging
import shutil
import textwrap
from collections.abc import Collection, Iterable, Iterator
from pathlib import Path, PurePosixPath
from typing import NamedTuple
from urllib.parse import urlparse
from urllib.request import url2pathname

from packaging.requirements import Requirement
from packaging.utils import canonicalize_name

try:
    from importlib import metadata
except ImportError:
    import importlib_metadata as metadata  # type: ignore[import-not-found,no-redef]

try:
    import tomllib
except ImportError:
    import tomli as tomllib  # type: ignore[import-not-found,no-redef]

from pipx.constants import MAN_SECTIONS, WINDOWS
from pipx.util import PipxError, run_subprocess

logger = logging.getLogger(__name__)


class VenvInspectInformation(NamedTuple):
    distributions: Collection[metadata.Distribution]
    env: dict[str, str]
    bin_path: Path
    man_path: Path


class VenvMetadata(NamedTuple):
    apps: list[str]
    app_paths: list[Path]
    apps_of_dependencies: list[str]
    app_paths_of_dependencies: dict[str, list[Path]]
    man_pages: list[str]
    man_paths: list[Path]
    man_pages_of_dependencies: list[str]
    man_paths_of_dependencies: dict[str, list[Path]]
    package_version: str
    python_version: str


def get_dist(package: str, distributions: Collection[metadata.Distribution]) -> metadata.Distribution | None:
    """Find matching distribution in the canonicalized sense."""
    for dist in distributions:
        if canonicalize_name(dist.metadata["name"]) == canonicalize_name(package):
            return dist
    return None


def get_package_dependencies(dist: metadata.Distribution, extras: set[str], env: dict[str, str]) -> list[Requirement]:
    eval_env = env.copy()
    # Add an empty extra to enable evaluation of non-extra markers
    if not extras:
        extras.add("")
    dependencies = []
    for req in map(Requirement, dist.requires or []):
        if not req.marker:
            dependencies.append(req)
        else:
            for extra in extras:
                eval_env["extra"] = extra
                if req.marker.evaluate(eval_env):
                    dependencies.append(req)
                    break

    return dependencies


def get_required_dependency_names(dist: metadata.Distribution, env: dict[str, str]) -> set[str]:
    """Canonical names ``dist`` depends on, including dependencies behind its declared extras.

    Considering every declared extra is deliberately conservative: a dependency that is only pulled
    in by an extra is still treated as required, so it is never reported as removable while the
    package declaring it stays installed.
    """
    extras = set(dist.metadata.get_all("Provides-Extra") or [])
    return {canonicalize_name(req.name) for req in get_package_dependencies(dist, extras, env)}


def list_not_required_packages(venv_python: Path) -> set[str]:
    """Canonical names of installed packages that no other installed package requires.

    Mirrors ``pip list --not-required`` from installed distribution metadata so every backend
    (including uv, which lacks the flag) computes the same result.
    """
    venv_sys_path, venv_env, _ = fetch_info_in_venv(venv_python)
    distributions = tuple(metadata.distributions(path=venv_sys_path))
    installed: set[str] = set()
    required: set[str] = set()
    for dist in distributions:
        if (name := dist.metadata["Name"]) is not None:
            installed.add(canonicalize_name(name))
        required |= get_required_dependency_names(dist, venv_env)
    return installed - required


def get_apps_from_entry_points(dist: metadata.Distribution, bin_path: Path):
    app_names = set()
    sections = {"console_scripts", "gui_scripts"}
    # "entry_points" entry in setup.py are found here
    for ep in dist.entry_points:
        if ep.group not in sections:
            continue
        if (bin_path / ep.name).exists():
            app_names.add(ep.name)
        if WINDOWS and (bin_path / (ep.name + ".exe")).exists():
            # WINDOWS adds .exe to entry_point name
            app_names.add(ep.name + ".exe")
    return app_names


def get_resources_from_dist_files(dist: metadata.Distribution, bin_path: Path, man_path: Path):
    app_names = set()
    man_names = set()
    # search installed files
    # "scripts" entry in setup.py is found here (test w/ awscli)
    for path in dist.files or []:
        # vast speedup by ignoring all paths not above distribution root dir
        #   (venv/bin or venv/Scripts is above distribution root)
        if Path(path).parts[0] != "..":
            continue

        dist_file_path = Path(str(dist.locate_file(path)))
        try:
            if dist_file_path.parent.samefile(bin_path):
                app_names.add(path.name)
            if dist_file_path.parent.name in MAN_SECTIONS and dist_file_path.parent.parent.samefile(man_path):
                man_names.add(str(Path(dist_file_path.parent.name) / path.name))
        except FileNotFoundError:
            pass
    return app_names, man_names


def get_resources_from_inst_files(dist: metadata.Distribution, bin_path: Path, man_path: Path):
    app_names = set()
    man_names = set()
    # not sure what is found here
    inst_files = dist.read_text("installed-files.txt") or ""
    for line in inst_files.splitlines():
        entry = line.split(",")[0]
        inst_file_path = Path(str(dist.locate_file(entry))).resolve()
        try:
            if inst_file_path.parent.samefile(bin_path):
                app_names.add(inst_file_path.name)
            if inst_file_path.parent.name in MAN_SECTIONS and inst_file_path.parent.parent.samefile(man_path):
                man_names.add(str(Path(inst_file_path.parent.name) / inst_file_path.name))
        except FileNotFoundError:
            pass
    return app_names, man_names


def get_resources(dist: metadata.Distribution, bin_path: Path, man_path: Path) -> tuple[list[str], list[str]]:
    app_names = set()
    man_names = set()
    app_names_ep = get_apps_from_entry_points(dist, bin_path)
    app_names_df, man_names_df = get_resources_from_dist_files(dist, bin_path, man_path)
    app_names_if, man_names_if = get_resources_from_inst_files(dist, bin_path, man_path)
    app_names = app_names_ep | app_names_df | app_names_if
    man_names = man_names_df | man_names_if | _get_man_pages_from_editable_project(dist, man_path)
    return sorted(app_names), sorted(man_names)


def _get_man_pages_from_editable_project(dist: metadata.Distribution, man_path: Path) -> set[str]:
    project_root = _get_editable_project_root(dist)
    if project_root is None:
        return set()

    man_names = set()
    for target_dir, source_paths in _iter_editable_project_data_files(project_root):
        man_section = _get_man_section_from_data_files_target(target_dir)
        if man_section is None:
            continue
        for source_path in source_paths:
            source = Path(source_path)
            if not source.is_absolute():
                source = project_root / source
            if not source.is_file():
                continue
            dest = man_path / man_section / source.name
            try:
                dest.parent.mkdir(parents=True, exist_ok=True)
                shutil.copy2(source, dest)
            except OSError as exc:
                logger.warning("Unable to copy editable man page %s to %s: %s", source, dest, exc)
                continue
            man_names.add(str(Path(man_section) / source.name))

    return man_names


def _get_editable_project_root(dist: metadata.Distribution) -> Path | None:
    direct_url = dist.read_text("direct_url.json")
    if not direct_url:
        return None
    try:
        data = json.loads(direct_url)
    except json.JSONDecodeError:
        return None

    dir_info = data.get("dir_info", {})
    if not isinstance(dir_info, dict) or not dir_info.get("editable"):
        return None

    url = data.get("url")
    if not isinstance(url, str):
        return None
    parsed_url = urlparse(url)
    if parsed_url.scheme != "file":
        return None

    url_path = f"//{parsed_url.netloc}{parsed_url.path}" if parsed_url.netloc else parsed_url.path
    return Path(url2pathname(url_path))


def _iter_editable_project_data_files(project_root: Path) -> Iterator[tuple[str, list[str]]]:
    yield from _iter_pyproject_data_files(project_root)
    yield from _iter_setup_cfg_data_files(project_root)
    yield from _iter_setup_py_data_files(project_root)


def _iter_pyproject_data_files(project_root: Path) -> Iterator[tuple[str, list[str]]]:
    pyproject = project_root / "pyproject.toml"
    try:
        with pyproject.open("rb") as file:
            pyproject_data = tomllib.load(file)
    except (OSError, tomllib.TOMLDecodeError):
        return

    data_files = pyproject_data.get("tool", {}).get("setuptools", {}).get("data-files", {})
    yield from _iter_normalized_data_files(data_files)


def _iter_setup_cfg_data_files(project_root: Path) -> Iterator[tuple[str, list[str]]]:
    setup_cfg = project_root / "setup.cfg"
    config = configparser.ConfigParser(interpolation=None)
    try:
        config.read(setup_cfg)
    except configparser.Error:
        return
    if not config.has_section("options.data_files"):
        return

    yield from _iter_normalized_data_files(dict(config.items("options.data_files")))


def _iter_setup_py_data_files(project_root: Path) -> Iterator[tuple[str, list[str]]]:
    setup_py = project_root / "setup.py"
    try:
        tree = ast.parse(setup_py.read_text(encoding="utf-8"))
    except (OSError, SyntaxError, UnicodeDecodeError):
        return

    for node in ast.walk(tree):
        if not isinstance(node, ast.Call) or not _is_setup_call(node):
            continue
        for keyword in node.keywords:
            if keyword.arg != "data_files":
                continue
            try:
                data_files = ast.literal_eval(keyword.value)
            except (ValueError, SyntaxError):
                return
            yield from _iter_normalized_data_files(data_files)
            return


def _is_setup_call(node: ast.Call) -> bool:
    return (isinstance(node.func, ast.Name) and node.func.id == "setup") or (
        isinstance(node.func, ast.Attribute) and node.func.attr == "setup"
    )


def _iter_normalized_data_files(data_files: object) -> Iterator[tuple[str, list[str]]]:
    items: Iterable[object]
    if isinstance(data_files, dict):
        items = data_files.items()
    elif isinstance(data_files, (list, tuple)):
        items = data_files
    else:
        return

    for item in items:
        if not isinstance(item, (list, tuple)) or len(item) != 2:
            continue
        target_dir, source_paths = item
        if not isinstance(target_dir, str):
            continue
        if isinstance(source_paths, str):
            yield target_dir, [line.strip() for line in source_paths.splitlines() if line.strip()]
        elif isinstance(source_paths, (list, tuple)) and all(isinstance(path, str) for path in source_paths):
            yield target_dir, list(source_paths)


def _get_man_section_from_data_files_target(target_dir: str) -> str | None:
    parts = PurePosixPath(target_dir.replace("\\", "/")).parts
    if len(parts) < 2 or parts[-2] != "man" or parts[-1] not in MAN_SECTIONS:
        return None
    return parts[-1]


def _dfs_package_resources(
    dist: metadata.Distribution,
    package_req: Requirement,
    venv_inspect_info: VenvInspectInformation,
    app_paths_of_dependencies: dict[str, list[Path]],
    man_paths_of_dependencies: dict[str, list[Path]],
    dep_visited: dict[str, bool] | None = None,
) -> tuple[dict[str, list[Path]], dict[str, list[Path]]]:
    if dep_visited is None:
        # Initialize: we have already visited root
        dep_visited = {canonicalize_name(package_req.name): True}

    dependencies = get_package_dependencies(dist, package_req.extras, venv_inspect_info.env)
    for dep_req in dependencies:
        dep_name = canonicalize_name(dep_req.name)
        if dep_name in dep_visited:
            # avoid infinite recursion, avoid duplicates in info
            continue

        dep_dist = get_dist(dep_req.name, venv_inspect_info.distributions)
        if dep_dist is None:
            raise PipxError(f"Pipx Internal Error: cannot find package {dep_req.name!r} metadata.")
        app_names, man_names = get_resources(dep_dist, venv_inspect_info.bin_path, venv_inspect_info.man_path)
        if app_names:
            app_paths_of_dependencies[dep_name] = [venv_inspect_info.bin_path / name for name in app_names]
        if man_names:
            man_paths_of_dependencies[dep_name] = [venv_inspect_info.man_path / name for name in man_names]
        # recursively search for more
        dep_visited[dep_name] = True
        app_paths_of_dependencies, man_paths_of_dependencies = _dfs_package_resources(
            dep_dist,
            dep_req,
            venv_inspect_info,
            app_paths_of_dependencies,
            man_paths_of_dependencies,
            dep_visited,
        )
    return app_paths_of_dependencies, man_paths_of_dependencies


def _windows_extra_app_paths(app_paths: list[Path]) -> list[Path]:
    # In Windows, editable package have additional files starting with the
    #   same name that are required to be in the same dir to run the app
    # Add "*-script.py", "*.exe.manifest" only to app_paths to make
    #   execution work; do not add them to apps to ensure they are not listed
    app_paths_output = app_paths.copy()
    for app_path in app_paths:
        win_app_path = app_path.parent / (app_path.stem + "-script.py")
        if win_app_path.exists():
            app_paths_output.append(win_app_path)
        win_app_path = app_path.parent / (app_path.stem + ".exe.manifest")
        if win_app_path.exists():
            app_paths_output.append(win_app_path)
    return app_paths_output


def fetch_info_in_venv(venv_python_path: Path) -> tuple[list[str], dict[str, str], str]:
    command_str = textwrap.dedent(
        """
        import json
        import os
        import platform
        import sys

        impl_ver = sys.implementation.version
        implementation_version = "{0.major}.{0.minor}.{0.micro}".format(impl_ver)
        if impl_ver.releaselevel != "final":
            implementation_version = "{}{}{}".format(
                implementation_version,
                impl_ver.releaselevel[0],
                impl_ver.serial,
            )

        sys_path = sys.path
        try:
            sys_path.remove("")
        except ValueError:
            pass

        print(
            json.dumps(
                {
                    "sys_path": sys_path,
                    "python_version": "{0.major}.{0.minor}.{0.micro}".format(sys.version_info),
                    "environment": {
                        "implementation_name": sys.implementation.name,
                        "implementation_version": implementation_version,
                        "os_name": os.name,
                        "platform_machine": platform.machine(),
                        "platform_release": platform.release(),
                        "platform_system": platform.system(),
                        "platform_version": platform.version(),
                        "python_full_version": platform.python_version(),
                        "platform_python_implementation": platform.python_implementation(),
                        "python_version": ".".join(platform.python_version_tuple()[:2]),
                        "sys_platform": sys.platform,
                    },
                }
            )
        )
        """
    )
    venv_info = json.loads(
        run_subprocess(
            [venv_python_path, "-c", command_str],
            capture_stderr=False,
            log_cmd_str="<fetch_info_in_venv commands>",
        ).stdout
    )
    return (
        venv_info["sys_path"],
        venv_info["environment"],
        f"Python {venv_info['python_version']}",
    )


def inspect_venv(
    root_package_name: str,
    root_package_extras: set[str],
    venv_bin_path: Path,
    venv_python_path: Path,
    venv_man_path: Path,
) -> VenvMetadata:
    app_paths_of_dependencies: dict[str, list[Path]] = {}
    apps_of_dependencies: list[str] = []
    man_paths_of_dependencies: dict[str, list[Path]] = {}
    man_pages_of_dependencies: list[str] = []

    root_req = Requirement(root_package_name)
    root_req.extras = root_package_extras

    (venv_sys_path, venv_env, venv_python_version) = fetch_info_in_venv(venv_python_path)

    # Collect the generator created from metadata.distributions()
    # (see `itertools.chain.from_iterable`) into a tuple because we
    # need to iterate over it multiple times in `_dfs_package_apps`.

    # Tuple is chosen over a list because the program only iterate over
    # the distributions and never modify it.
    distributions = tuple(metadata.distributions(path=venv_sys_path))

    venv_inspect_info = VenvInspectInformation(
        bin_path=venv_bin_path,
        man_path=venv_man_path,
        env=venv_env,
        distributions=distributions,
    )

    root_dist = get_dist(root_req.name, venv_inspect_info.distributions)
    if root_dist is None:
        raise PipxError(f"Pipx Internal Error: cannot find package {root_req.name!r} metadata.")
    app_paths_of_dependencies, man_paths_of_dependencies = _dfs_package_resources(
        root_dist,
        root_req,
        venv_inspect_info,
        app_paths_of_dependencies,
        man_paths_of_dependencies,
    )

    apps, man_pages = get_resources(root_dist, venv_bin_path, venv_man_path)
    app_paths = [venv_bin_path / app for app in apps]
    man_paths = [venv_man_path / man_page for man_page in man_pages]
    if WINDOWS:
        app_paths = _windows_extra_app_paths(app_paths)

    for dep in app_paths_of_dependencies:
        apps_of_dependencies += [dep_path.name for dep_path in app_paths_of_dependencies[dep]]
        if WINDOWS:
            app_paths_of_dependencies[dep] = _windows_extra_app_paths(app_paths_of_dependencies[dep])
    for dep in man_paths_of_dependencies:
        man_pages_of_dependencies += [
            str(Path(dep_path.parent.name) / dep_path.name) for dep_path in man_paths_of_dependencies[dep]
        ]

    return VenvMetadata(
        apps=apps,
        app_paths=app_paths,
        apps_of_dependencies=apps_of_dependencies,
        app_paths_of_dependencies=app_paths_of_dependencies,
        man_pages=man_pages,
        man_paths=man_paths,
        man_pages_of_dependencies=man_pages_of_dependencies,
        man_paths_of_dependencies=man_paths_of_dependencies,
        package_version=root_dist.version,
        python_version=venv_python_version,
    )



# Valid package specifiers for pipx:
#   PEP508-compliant
#   git+<URL>
#   <URL>
#   <pypi_package_name>
#   <pypi_package_name><version_specifier>
#   <local_path>

import logging
import re
import urllib.parse
from dataclasses import dataclass
from pathlib import Path

from packaging.requirements import InvalidRequirement, Requirement
from packaging.specifiers import SpecifierSet
from packaging.utils import canonicalize_name

from pipx.emojis import hazard
from pipx.util import PipxError, pipx_wrap

logger = logging.getLogger(__name__)

ARCHIVE_EXTENSIONS = (".whl", ".tar.gz", ".zip")


@dataclass(frozen=True)
class ParsedPackage:
    valid_pep508: Requirement | None
    valid_url: str | None
    valid_local_path: str | None


def _split_path_extras(package_spec: str) -> tuple[str, str]:
    """Returns (path, extras_string)"""
    package_spec_extras_re = re.search(r"(.+)(\[.+\])", package_spec)
    if package_spec_extras_re:
        return (package_spec_extras_re.group(1), package_spec_extras_re.group(2))
    else:
        return (package_spec, "")


def _check_package_path(package_path: str) -> tuple[Path, bool]:
    pkg_path = Path(package_path)
    pkg_path_exists = pkg_path.exists()

    return (pkg_path, pkg_path_exists)


def _parse_specifier(package_spec: str) -> ParsedPackage:
    """Parse package_spec as would be given to pipx"""
    # If package_spec is valid pypi name, pip will always treat it as a
    #       pypi package, not checking for local path.
    #       We replicate pypi precedence here (only non-valid-pypi names
    #       initiate check for local path, e.g. './package-name')
    valid_pep508 = None
    valid_url = None
    valid_local_path = None

    try:
        package_req = Requirement(package_spec)
    except InvalidRequirement:
        # not a valid PEP508 package specification
        pass
    else:
        # valid PEP508 package specification
        valid_pep508 = package_req

    if valid_pep508 and package_req.name.endswith(ARCHIVE_EXTENSIONS):
        # It might be a local archive
        (package_path, package_path_exists) = _check_package_path(package_req.name)

        if package_path_exists:
            valid_local_path = str(package_path.resolve())
        else:
            raise PipxError(f"{package_path} does not exist")

    # If this looks like a URL, treat it as such.
    if not valid_pep508:
        parsed_url = urllib.parse.urlsplit(package_spec)
        if parsed_url.scheme and parsed_url.netloc:
            valid_url = package_spec

    # Treat the input as a local path if it does not look like a PEP 508
    # specifier nor a URL. In this case we want to split out the extra part.
    if not valid_pep508 and not valid_url:
        (package_path_str, package_extras_str) = _split_path_extras(package_spec)
        (package_path, package_path_exists) = _check_package_path(package_path_str)
        if package_path_exists:
            valid_local_path = str(package_path.resolve()) + package_extras_str

    if not valid_pep508 and not valid_url and not valid_local_path:
        raise PipxError(f"Unable to parse package spec: {package_spec}")

    if valid_pep508 and valid_local_path:
        # It is a valid local path without "./"
        # Use valid_local_path
        valid_pep508 = None

    return ParsedPackage(
        valid_pep508=valid_pep508,
        valid_url=valid_url,
        valid_local_path=valid_local_path,
    )


def package_or_url_from_pep508(requirement: Requirement, remove_version_specifiers: bool = False) -> str:
    requirement.marker = None
    requirement.name = canonicalize_name(requirement.name)
    if remove_version_specifiers:
        requirement.specifier = SpecifierSet("")
    return str(requirement)


def _parsed_package_to_package_or_url(parsed_package: ParsedPackage, remove_version_specifiers: bool) -> str:
    if parsed_package.valid_pep508 is not None:
        if parsed_package.valid_pep508.marker is not None:
            logger.warning(
                pipx_wrap(
                    f"""
                    {hazard}  Ignoring environment markers
                    ({parsed_package.valid_pep508.marker}) in package
                    specification. Use pipx options to specify this type of
                    information.
                    """,
                    subsequent_indent=" " * 4,
                )
            )
        package_or_url = package_or_url_from_pep508(
            parsed_package.valid_pep508,
            remove_version_specifiers=remove_version_specifiers,
        )
    elif parsed_package.valid_url is not None:
        package_or_url = parsed_package.valid_url
    elif parsed_package.valid_local_path is not None:
        package_or_url = parsed_package.valid_local_path

    logger.info(f"cleaned package spec: {package_or_url}")
    return package_or_url


def parse_specifier_for_install(package_spec: str, pip_args: list[str]) -> tuple[str, list[str]]:
    """Return package_or_url and pip_args suitable for pip install

    Specifically:
    * Strip any markers (e.g. python_version > "3.4")
    * Ensure --editable is removed for any package_spec not a local path
    * Convert local paths to absolute paths
    """
    parsed_package = _parse_specifier(package_spec)
    package_or_url = _parsed_package_to_package_or_url(parsed_package, remove_version_specifiers=False)
    if "--editable" in pip_args and not parsed_package.valid_local_path:
        logger.warning(
            pipx_wrap(
                f"""
                {hazard}  Ignoring --editable install option. pipx disallows it
                for anything but a local path, to avoid having to create a new
                src/ directory.
                """,
                subsequent_indent=" " * 4,
            )
        )
        pip_args.remove("--editable")

    for index, option in enumerate(pip_args):
        if not option.startswith(("-c", "--constraint")):
            continue

        if option in ("-c", "--constraint"):
            argument_index = index + 1
            if argument_index < len(pip_args) and not urllib.parse.urlsplit(pip_args[argument_index]).scheme:
                pip_args[argument_index] = str(Path(pip_args[argument_index]).expanduser().resolve())

        elif (option_list := option.split("=", maxsplit=1)) and len(option_list) == 2:
            key, value = option_list
            if not urllib.parse.urlsplit(value).scheme:
                pip_args[index] = f"{key}={Path(value).expanduser().resolve()}"

        break

    return package_or_url, pip_args


def parse_specifier_for_metadata(package_spec: str) -> str:
    """Return package_or_url suitable for pipx metadata

    Specifically:
    * Strip any markers (e.g. python_version > 3.4)
    * Convert local paths to absolute paths
    """
    parsed_package = _parse_specifier(package_spec)
    return _parsed_package_to_package_or_url(parsed_package, remove_version_specifiers=False)


def parse_specifier_for_upgrade(package_spec: str) -> str:
    """Return package_or_url suitable for pip upgrade

    Specifically:
    * Strip any version specifiers (e.g. package == 1.5.4)
    * Strip any markers (e.g. python_version > 3.4)
    * Convert local paths to absolute paths
    """
    parsed_package = _parse_specifier(package_spec)
    return _parsed_package_to_package_or_url(parsed_package, remove_version_specifiers=True)


def get_extras(package_spec: str) -> set[str]:
    parsed_package = _parse_specifier(package_spec)
    if parsed_package.valid_pep508 and parsed_package.valid_pep508.extras is not None:
        return parsed_package.valid_pep508.extras
    elif parsed_package.valid_local_path:
        (_, package_extras_str) = _split_path_extras(parsed_package.valid_local_path)
        return Requirement("notapackage" + package_extras_str).extras

    return set()


def valid_pypi_name(package_spec: str) -> str | None:
    try:
        package_req = Requirement(package_spec)
    except InvalidRequirement:
        # not a valid PEP508 package specification
        return None

    if package_req.url or package_req.name.endswith(ARCHIVE_EXTENSIONS):
        # package name supplied by user might not match package found in URL,
        # also if package name ends with archive extension, it might be a local archive file,
        #   so force package name determination the long way
        return None

    return canonicalize_name(package_req.name)


def fix_package_name(package_or_url: str, package_name: str) -> str:
    try:
        package_req = Requirement(package_or_url)
    except InvalidRequirement:
        # not a valid PEP508 package specification
        return package_or_url

    if package_req.name.endswith(ARCHIVE_EXTENSIONS):
        return str(package_req)

    if canonicalize_name(package_req.name) != canonicalize_name(package_name):
        logger.warning(
            pipx_wrap(
                f"""
                {hazard}  Name supplied in package specifier was
                {package_req.name!r} but package found has name {package_name!r}.
                Using {package_name!r}.
                """,
                subsequent_indent=" " * 4,
            )
        )
    package_req.name = package_name

    return str(package_req)



import logging
import os
from pathlib import Path

from platformdirs import user_cache_path, user_data_path, user_log_path

from pipx.constants import LINUX, WINDOWS
from pipx.emojis import hazard, strtobool
from pipx.wrap import pipx_wrap

if LINUX:
    DEFAULT_PIPX_HOME = Path(user_data_path("pipx"))
    FALLBACK_PIPX_HOMES = [Path.home() / ".local/pipx"]
elif WINDOWS:
    DEFAULT_PIPX_HOME = Path.home() / "pipx"
    FALLBACK_PIPX_HOMES = [Path.home() / ".local/pipx", Path(user_data_path("pipx"))]
else:
    DEFAULT_PIPX_HOME = Path.home() / ".local/pipx"
    FALLBACK_PIPX_HOMES = [Path(user_data_path("pipx"))]

DEFAULT_PIPX_BIN_DIR = Path.home() / ".local/bin"
DEFAULT_PIPX_MAN_DIR = Path.home() / ".local/share/man"
DEFAULT_PIPX_GLOBAL_HOME = Path("/opt/pipx")
DEFAULT_PIPX_GLOBAL_BIN_DIR = Path("/usr/local/bin")
DEFAULT_PIPX_GLOBAL_MAN_DIR = Path("/usr/local/share/man")

# Overrides for testing
OVERRIDE_PIPX_HOME = None
OVERRIDE_PIPX_BIN_DIR = None
OVERRIDE_PIPX_MAN_DIR = None
OVERRIDE_PIPX_SHARED_LIBS = None
OVERRIDE_PIPX_GLOBAL_HOME = None
OVERRIDE_PIPX_GLOBAL_BIN_DIR = None
OVERRIDE_PIPX_GLOBAL_MAN_DIR = None

logger = logging.getLogger(__name__)


def get_expanded_environ(env_name: str) -> Path | None:
    val = os.environ.get(env_name)
    if val is not None:
        return Path(val).expanduser().resolve()
    return val


class _PathContext:
    _base_home: Path | None
    _default_home: Path
    _base_bin: Path | None
    _default_bin: Path
    _base_man: Path | None
    _default_man: Path
    _default_log: Path
    _default_cache: Path
    _default_trash: Path
    _base_shared_libs: Path | None
    _fallback_home: Path | None
    _home_exists: bool
    log_file: Path | None = None

    def __init__(self):
        self.make_local()

    @property
    def venvs(self) -> Path:
        return self.home / "venvs"

    @property
    def logs(self) -> Path:
        if self._home_exists or not LINUX:
            return self.home / "logs"
        return self._default_log

    @property
    def trash(self) -> Path:
        if self._home_exists:
            return self.home / ".trash"
        return self._default_trash

    @property
    def venv_cache(self) -> Path:
        if self._home_exists or not LINUX:
            return self.home / ".cache"
        return self._default_cache

    @property
    def bin_dir(self) -> Path:
        return (self._base_bin or self._default_bin).resolve()

    @property
    def man_dir(self) -> Path:
        return (self._base_man or self._default_man).resolve()

    @property
    def home(self) -> Path:
        if self._base_home:
            home = Path(self._base_home)
        elif self._fallback_home:
            home = self._fallback_home
        else:
            home = self._default_home
        return home.resolve()

    @property
    def shared_libs(self) -> Path:
        return (self._base_shared_libs or self.home / "shared").resolve()

    def make_local(self) -> None:
        self._base_home = OVERRIDE_PIPX_HOME or get_expanded_environ("PIPX_HOME")  # type: ignore[redundant-expr]
        self._default_home = DEFAULT_PIPX_HOME
        self._base_bin = OVERRIDE_PIPX_BIN_DIR or get_expanded_environ("PIPX_BIN_DIR")  # type: ignore[redundant-expr]
        self._default_bin = DEFAULT_PIPX_BIN_DIR
        self._base_man = OVERRIDE_PIPX_MAN_DIR or get_expanded_environ("PIPX_MAN_DIR")  # type: ignore[redundant-expr]
        self._default_man = DEFAULT_PIPX_MAN_DIR
        self._base_shared_libs = OVERRIDE_PIPX_SHARED_LIBS or get_expanded_environ("PIPX_SHARED_LIBS")  # type: ignore[redundant-expr]
        self._default_log = Path(user_log_path("pipx"))
        self._default_cache = Path(user_cache_path("pipx"))
        self._default_trash = self._default_home / "trash"
        self._fallback_home = next(iter([fallback for fallback in FALLBACK_PIPX_HOMES if fallback.exists()]), None)
        self._home_exists = self._base_home is not None or any(fallback.exists() for fallback in FALLBACK_PIPX_HOMES)

    def make_global(self) -> None:
        self._base_home = OVERRIDE_PIPX_GLOBAL_HOME or get_expanded_environ("PIPX_GLOBAL_HOME")  # type: ignore[redundant-expr]
        self._default_home = DEFAULT_PIPX_GLOBAL_HOME
        self._base_bin = OVERRIDE_PIPX_GLOBAL_BIN_DIR or get_expanded_environ("PIPX_GLOBAL_BIN_DIR")  # type: ignore[redundant-expr]
        self._default_bin = DEFAULT_PIPX_GLOBAL_BIN_DIR
        self._base_man = OVERRIDE_PIPX_GLOBAL_MAN_DIR or get_expanded_environ("PIPX_GLOBAL_MAN_DIR")  # type: ignore[redundant-expr]
        self._default_man = DEFAULT_PIPX_GLOBAL_MAN_DIR
        self._default_log = self._default_home / "logs"
        self._default_cache = self._default_home / ".cache"
        self._default_trash = self._default_home / "trash"
        self._base_shared_libs = None
        self._fallback_home = None
        self._home_exists = self._base_home is not None

    @property
    def standalone_python_cachedir(self) -> Path:
        return self.home / "py"

    @property
    def allow_spaces_in_home_path(self) -> bool:
        return strtobool(os.getenv("PIPX_HOME_ALLOW_SPACE", "0"))

    def log_warnings(self):
        if " " in str(self.home) and not self.allow_spaces_in_home_path:
            logger.warning(
                pipx_wrap(
                    (
                        f"{hazard} Found a space in the pipx home path. We heavily discourage this, due to "
                        "multiple incompatibilities. Please check our docs for more information on this, "
                        "as well as some pointers on how to migrate to a different home path."
                    ),
                    subsequent_indent=" " * 4,
                )
            )
            logger.warning(
                pipx_wrap(
                    (f"{hazard} To see your PIPX_HOME dir: pipx environment --value PIPX_HOME"),
                    subsequent_indent=" " * 4,
                )
            )
            logger.warning(
                pipx_wrap(
                    (f"{hazard} Most likely fix on macOS: mv ~/Library/Application\\ Support/pipx ~/.local/"),
                    subsequent_indent=" " * 4,
                )
            )

        fallback_home_exists = self._fallback_home is not None and self._fallback_home.exists()
        specific_home_exists = self.home != self._fallback_home
        if fallback_home_exists and specific_home_exists:
            logger.info(
                pipx_wrap(
                    (
                        f"Both a specific pipx home folder ({self.home}) and the fallback "
                        f"pipx home folder ({self._fallback_home}) exist. If you are done migrating from the"
                        "fallback to the new location, it is safe to delete the fallback location."
                    ),
                    subsequent_indent=" " * 4,
                )
            )


ctx = _PathContext()
ctx.log_warnings()



import sys
from collections.abc import Sequence
from pathlib import Path
from tempfile import mkdtemp

from packaging.utils import canonicalize_name

from pipx.commands.common import add_suffix
from pipx.commands.inject import inject_dep
from pipx.commands.install import install
from pipx.commands.uninstall import _get_venv_resource_paths
from pipx.constants import (
    EXIT_CODE_OK,
    EXIT_CODE_REINSTALL_INVALID_PYTHON,
    EXIT_CODE_REINSTALL_VENV_NONEXISTENT,
    MAN_SECTIONS,
    ExitCode,
)
from pipx.emojis import error, sleep, stars
from pipx.util import PipxError, rmdir, safe_unlink
from pipx.venv import Venv, VenvContainer


def _create_reinstall_backup(venv_dir: Path) -> Path:
    backup_dir = Path(mkdtemp(prefix=f".{venv_dir.name}-", suffix="-pipx-reinstall", dir=venv_dir.parent))
    backup_dir.rmdir()
    venv_dir.rename(backup_dir)
    return backup_dir


def _restore_reinstall_backup(venv_dir: Path, restore_venv_dir: Path, backup_dir: Path) -> None:
    rmdir(venv_dir)
    backup_dir.rename(restore_venv_dir)


def _get_reinstall_resource_paths(venv: Venv, local_bin_dir: Path, local_man_dir: Path) -> set[Path]:
    resource_paths = _get_venv_resource_paths("app", venv, venv.bin_path, local_bin_dir)
    for man_section in MAN_SECTIONS:
        resource_paths |= _get_venv_resource_paths("man", venv, venv.man_path / man_section, local_man_dir / man_section)
    return resource_paths


def _get_expected_reinstall_resource_paths(venv: Venv, local_bin_dir: Path, local_man_dir: Path) -> set[Path]:
    resource_paths: set[Path] = set()
    for package_info in venv.package_metadata.values():
        if package_info.include_apps:
            for app_path in package_info.app_paths:
                resource_paths.add(local_bin_dir / add_suffix(app_path.name, package_info.suffix))
            for man_path in package_info.man_paths:
                resource_paths.add(local_man_dir / man_path.parent.name / man_path.name)
        if package_info.include_dependencies:
            for app_paths in package_info.app_paths_of_dependencies.values():
                for app_path in app_paths:
                    resource_paths.add(local_bin_dir / add_suffix(app_path.name, package_info.suffix))
            for man_paths in package_info.man_paths_of_dependencies.values():
                for man_path in man_paths:
                    resource_paths.add(local_man_dir / man_path.parent.name / man_path.name)
    return resource_paths


def _remove_stale_reinstall_resources(resource_paths: set[Path]) -> None:
    for path in sorted(resource_paths):
        try:
            safe_unlink(path)
            if path.is_symlink():
                path.unlink()
        except FileNotFoundError:
            pass


def reinstall(
    *,
    venv_dir: Path,
    local_bin_dir: Path,
    local_man_dir: Path,
    python: str,
    verbose: bool,
    force_reinstall_shared_libs: bool = False,
    python_flag_passed: bool = False,
    backend: str | None = None,
    env_backend: str | None = None,
) -> ExitCode:
    """Returns pipx exit code."""
    if not venv_dir.exists():
        print(f"Nothing to reinstall for {venv_dir.name} {sleep}")
        return EXIT_CODE_REINSTALL_VENV_NONEXISTENT

    try:
        Path(python).relative_to(venv_dir)
    except ValueError:
        pass
    else:
        print(
            f"{error} Error, the python executable would be deleted!",
            "Change it using the --python option or PIPX_DEFAULT_PYTHON environment variable.",
        )
        return EXIT_CODE_REINSTALL_INVALID_PYTHON

    venv = Venv(venv_dir, verbose=verbose, backend=backend, env_backend=env_backend)
    venv.check_upgrade_shared_libs(
        pip_args=venv.pipx_metadata.main_package.pip_args, verbose=verbose, force_upgrade=force_reinstall_shared_libs
    )

    if venv.pipx_metadata.main_package.package_or_url is not None:
        package_or_url = venv.pipx_metadata.main_package.package_or_url
    else:
        package_or_url = venv.main_package_name

    if venv.pipx_metadata.main_package.pinned:
        raise PipxError(f"{error} Package {venv_dir} is pinned. Run `pipx unpin {venv_dir.name}` to unpin it first.")

    old_resource_paths = _get_reinstall_resource_paths(venv, local_bin_dir, local_man_dir)
    original_venv_dir = venv_dir
    reinstall_backup_dir = _create_reinstall_backup(venv_dir)
    print(f"uninstalled {venv.name}! {stars}")

    # in case legacy original dir name
    venv_dir = venv_dir.with_name(canonicalize_name(venv_dir.name))

    try:
        # install main package first
        install(
            venv_dir,
            [venv.main_package_name],
            [package_or_url],
            local_bin_dir,
            local_man_dir,
            python,
            venv.pipx_metadata.main_package.pip_args,
            venv.pipx_metadata.venv_args,
            verbose,
            force=True,
            reinstall=True,
            include_dependencies=venv.pipx_metadata.main_package.include_dependencies,
            preinstall_packages=[],
            suffix=venv.pipx_metadata.main_package.suffix,
            python_flag_passed=python_flag_passed,
            backend=backend or venv.pipx_metadata.backend,
            env_backend=env_backend,
        )

        # now install injected packages
        for injected_name, injected_package in venv.pipx_metadata.injected_packages.items():
            if injected_package.package_or_url is None:
                # This should never happen, but package_or_url is type
                #   Optional[str] so mypy thinks it could be None
                raise PipxError(f"Internal Error injecting package {injected_package} into {venv.name}")
            inject_dep(
                venv_dir,
                injected_name,
                injected_package.package_or_url,
                injected_package.pip_args,
                verbose=verbose,
                include_apps=injected_package.include_apps,
                include_dependencies=injected_package.include_dependencies,
                force=True,
                backend=backend or venv.pipx_metadata.backend,
                env_backend=env_backend,
            )

        new_resource_paths = _get_expected_reinstall_resource_paths(
            Venv(venv_dir, verbose=verbose), local_bin_dir, local_man_dir
        )
        _remove_stale_reinstall_resources(old_resource_paths - new_resource_paths)
    except (Exception, KeyboardInterrupt):
        _restore_reinstall_backup(venv_dir, original_venv_dir, reinstall_backup_dir)
        print(f"{error} Reinstall failed; restored {venv.name}.", file=sys.stderr)
        raise
    else:
        rmdir(reinstall_backup_dir)

    # Any failure to install will raise PipxError, otherwise success
    return EXIT_CODE_OK


def reinstall_all(
    venv_container: VenvContainer,
    local_bin_dir: Path,
    local_man_dir: Path,
    python: str,
    verbose: bool,
    *,
    skip: Sequence[str],
    python_flag_passed: bool = False,
    backend: str | None = None,
    env_backend: str | None = None,
) -> ExitCode:
    """Returns pipx exit code."""
    failed: list[str] = []
    reinstalled: list[str] = []

    # iterate on all packages and reinstall them
    # for the first one, we also trigger
    # a reinstall of shared libs beforehand
    first_reinstall = True
    for venv_dir in venv_container.iter_venv_dirs():
        if venv_dir.name in skip:
            continue
        try:
            reinstall(
                venv_dir=venv_dir,
                local_bin_dir=local_bin_dir,
                local_man_dir=local_man_dir,
                python=python,
                verbose=verbose,
                force_reinstall_shared_libs=first_reinstall,
                python_flag_passed=python_flag_passed,
                backend=backend,
                env_backend=env_backend,
            )
        except PipxError as e:
            print(e, file=sys.stderr)
            failed.append(venv_dir.name)
        else:
            first_reinstall = False
            reinstalled.append(venv_dir.name)
    if len(reinstalled) == 0:
        print(f"No packages reinstalled after running 'pipx reinstall-all' {sleep}")
    if len(failed) > 0:
        raise PipxError(f"The following package(s) failed to reinstall: {', '.join(failed)}")
    # Any failure to install will raise PipxError, otherwise success
    return EXIT_CODE_OK


__all__ = [
    "reinstall",
    "reinstall_all",
]



import logging
from pathlib import Path
from shutil import which
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable

from pipx.commands.common import (
    add_suffix,
    can_symlink,
    get_exposed_man_paths_for_package,
    get_exposed_paths_for_package,
)
from pipx.constants import (
    EXIT_CODE_OK,
    EXIT_CODE_UNINSTALL_ERROR,
    EXIT_CODE_UNINSTALL_VENV_NONEXISTENT,
    MAN_SECTIONS,
    ExitCode,
)
from pipx.emojis import hazard, sleep, stars
from pipx.pipx_metadata_file import PackageInfo
from pipx.util import rmdir, safe_unlink
from pipx.venv import Venv, VenvContainer
from pipx.venv_inspect import VenvMetadata

logger = logging.getLogger(__name__)


def _venv_metadata_to_package_info(
    venv_metadata: VenvMetadata,
    package_name: str,
    package_or_url: str = "",
    pip_args: list[str] | None = None,
    include_apps: bool = True,
    include_dependencies: bool = False,
    suffix: str = "",
) -> PackageInfo:
    if pip_args is None:
        pip_args = []

    return PackageInfo(
        package=package_name,
        package_or_url=package_or_url,
        pip_args=pip_args,
        include_apps=include_apps,
        include_dependencies=include_dependencies,
        apps=venv_metadata.apps,
        app_paths=venv_metadata.app_paths,
        apps_of_dependencies=venv_metadata.apps_of_dependencies,
        app_paths_of_dependencies=venv_metadata.app_paths_of_dependencies,
        man_pages=venv_metadata.man_pages,
        man_paths=venv_metadata.man_paths,
        man_pages_of_dependencies=venv_metadata.man_pages_of_dependencies,
        man_paths_of_dependencies=venv_metadata.man_paths_of_dependencies,
        package_version=venv_metadata.package_version,
        suffix=suffix,
    )


def _get_package_bin_dir_app_paths(
    venv: Venv, package_info: PackageInfo, venv_bin_path: Path, local_bin_dir: Path
) -> set[Path]:
    suffix = package_info.suffix
    apps = []
    if package_info.include_apps:
        apps += package_info.apps
    if package_info.include_dependencies:
        apps += package_info.apps_of_dependencies
    return get_exposed_paths_for_package(venv_bin_path, local_bin_dir, [add_suffix(app, suffix) for app in apps])


def _get_package_man_paths(venv: Venv, package_info: PackageInfo, venv_man_path: Path, local_man_dir: Path) -> set[Path]:
    man_pages = []
    if package_info.include_apps:
        man_pages += package_info.man_pages
    if package_info.include_dependencies:
        man_pages += package_info.man_pages_of_dependencies
    return get_exposed_man_paths_for_package(venv_man_path, local_man_dir, man_pages)


def _get_venv_resource_paths(
    resource_type: str, venv: Venv, venv_resource_path: Path, local_resource_dir: Path
) -> set[Path]:
    resource_paths = set()
    assert resource_type in ("app", "man"), "invalid resource type"
    get_package_resource_paths: Callable[[Venv, PackageInfo, Path, Path], set[Path]]
    get_package_resource_paths = {
        "app": _get_package_bin_dir_app_paths,
        "man": _get_package_man_paths,
    }[resource_type]

    if venv.pipx_metadata.main_package.package is not None:
        # Valid metadata for venv
        for package_info in venv.package_metadata.values():
            resource_paths |= get_package_resource_paths(venv, package_info, venv_resource_path, local_resource_dir)
    elif venv.python_path.is_file():
        # No metadata from pipx_metadata.json, but valid python interpreter.
        # In pre-metadata-pipx venv.root.name is name of main package
        # In pre-metadata-pipx there is no suffix
        # We make the conservative assumptions: no injected packages,
        # not include_dependencies.  Other PackageInfo fields are irrelevant
        # here.
        venv_metadata = venv.get_venv_metadata_for_package(venv.root.name, set())
        main_package_info = _venv_metadata_to_package_info(venv_metadata, venv.root.name)
        resource_paths = get_package_resource_paths(venv, main_package_info, venv_resource_path, local_resource_dir)
    else:
        # No metadata and no valid python interpreter.
        # We'll take our best guess on what to uninstall here based on symlink
        # location for symlink-capable systems.
        # The heuristic here is any symlink in ~/.local/bin pointing to
        # .local/share/pipx/venvs/VENV_NAME/{bin,Scripts} should be uninstalled.

        # For non-symlink systems we give up and return an empty set.
        if not local_resource_dir.is_dir() or not can_symlink(local_resource_dir):
            return set()

        resource_paths = get_exposed_paths_for_package(venv_resource_path, local_resource_dir)

    return resource_paths


def uninstall(venv_dir: Path, local_bin_dir: Path, local_man_dir: Path, verbose: bool) -> ExitCode:
    """Uninstall entire venv_dir, including main package and all injected
    packages.

    Returns pipx exit code.
    """
    if not venv_dir.exists():
        print(f"Nothing to uninstall for {venv_dir.name} {sleep}")
        app = which(venv_dir.name)
        if app:
            print(f"{hazard}  Note: '{app}' still exists on your system and is on your PATH")
        return EXIT_CODE_UNINSTALL_VENV_NONEXISTENT

    venv = Venv(venv_dir, verbose=verbose)

    bin_dir_app_paths = _get_venv_resource_paths("app", venv, venv.bin_path, local_bin_dir)
    man_dir_paths = set()
    for man_section in MAN_SECTIONS:
        man_dir_paths |= _get_venv_resource_paths("man", venv, venv.man_path / man_section, local_man_dir / man_section)

    for path in bin_dir_app_paths | man_dir_paths:
        try:
            safe_unlink(path)
        except FileNotFoundError:
            logger.info(f"tried to remove but couldn't find {path}")
        else:
            logger.info(f"removed file {path}")

    rmdir(venv_dir)
    print(f"uninstalled {venv.name}! {stars}")
    return EXIT_CODE_OK


def uninstall_all(
    venv_container: VenvContainer,
    local_bin_dir: Path,
    local_man_dir: Path,
    verbose: bool,
) -> ExitCode:
    """Returns pipx exit code."""
    all_success = True
    for venv_dir in venv_container.iter_venv_dirs():
        return_val = uninstall(venv_dir, local_bin_dir, local_man_dir, verbose)
        all_success &= return_val == 0

    return EXIT_CODE_OK if all_success else EXIT_CODE_UNINSTALL_ERROR



from __future__ import annotations

import json
import logging
import re
import shutil
import subprocess
from functools import cache
from pathlib import Path
from typing import TYPE_CHECKING, Final

from packaging.version import InvalidVersion, Version

from pipx.animate import animate
from pipx.backends._base import UV, Backend
from pipx.util import (
    PipxError,
    run_subprocess,
    subprocess_post_check,
    subprocess_post_check_handle_pip_error,
)
from pipx.venv_inspect import list_not_required_packages

if TYPE_CHECKING:
    from collections.abc import Callable
    from subprocess import CompletedProcess

_LOGGER: Final[logging.Logger] = logging.getLogger(__name__)


# Imported into a temporary, then assigned through Final and deleted so mypy
# sees one Final binding (it rejects redefinition across try/except branches).
try:
    from uv import find_uv_bin as _uv_bin_from_extra  # type: ignore[import-not-found]
except ImportError:
    _uv_bin_from_extra = None
_FIND_UV_BIN_FROM_EXTRA: Final[Callable[[], str] | None] = _uv_bin_from_extra
del _uv_bin_from_extra
_MIN_UV_VERSION: Final[Version] = Version("0.4.0")
_VERSION_RE: Final[re.Pattern[str]] = re.compile(
    r"""
    uv \s+        # the literal "uv " prefix from `uv --version`
    (\S+)         # capture the version token; PEP 440 parser validates it
    """,
    re.VERBOSE,
)
# Stripping VIRTUAL_ENV stops uv from auto-targeting an active venv when no
# ``--python`` flag is passed.
_UV_ENV_OVERRIDES: Final[dict[str, str | None]] = {"VIRTUAL_ENV": None, "UV_NO_PROGRESS": "1"}


class UvBackend(Backend):
    name = UV

    def __init__(self) -> None:
        self._binary = resolve_uv_binary()
        _, self._source = find_uv_binary()
        version = _check_uv_version(self._binary)
        _LOGGER.info(f"using {self._source} uv {version} from {self._binary}")

    def needs_shared_libs(self) -> bool:
        return False

    def upgrade_packaging_libraries(self, venv_python: Path, pip_args: list[str], *, verbose: bool) -> None:
        del venv_python, pip_args, verbose  # uv venvs ship no pip to upgrade.

    def create_venv(
        self,
        root: Path,
        *,
        python: str,
        venv_args: list[str],
        pip_args: list[str],
        include_pip: bool,
        verbose: bool,
    ) -> None:
        del pip_args  # uv venv has no pip to seed.
        if include_pip:
            raise PipxError(
                "The uv backend cannot create a virtual environment with pip preinstalled.\n"
                "Reinstall the package with `--backend pip` (or unset PIPX_DEFAULT_BACKEND)."
            )
        cmd: list[str | Path] = [self._binary, "venv", "--python", python, *venv_args]
        cmd.append("--verbose" if verbose else "--quiet")
        cmd.append(str(root))
        with animate("creating virtual environment", not verbose):
            process = run_subprocess(cmd, run_dir=str(root), env_overrides=_UV_ENV_OVERRIDES)
        subprocess_post_check(process)

    def install(
        self,
        *,
        venv_root: Path,
        venv_python: Path,
        requirements: list[str],
        pip_args: list[str],
        no_deps: bool = False,
        upgrade: bool = False,
        log_pip_errors: bool = True,
        verbose: bool = False,
    ) -> CompletedProcess[str]:
        cmd = self._uv_pip_command("install", venv_python, verbose=verbose)
        if upgrade:
            cmd.append("--upgrade")
        if no_deps:
            cmd.append("--no-deps")
        cmd += [*_strip_pip_quiet_flags(pip_args), *requirements]
        process = run_subprocess(
            cmd,
            run_dir=str(venv_root),
            log_stdout=not log_pip_errors,
            log_stderr=not log_pip_errors,
            env_overrides=_UV_ENV_OVERRIDES,
        )
        if log_pip_errors:
            subprocess_post_check_handle_pip_error(process, tool_name="uv")
        return process

    def uninstall(
        self,
        *,
        venv_root: Path,
        venv_python: Path,
        package: str,
        verbose: bool,
    ) -> CompletedProcess[str]:
        cmd = self._uv_pip_command("uninstall", venv_python, verbose=verbose)
        cmd.append(package)
        process = run_subprocess(cmd, run_dir=str(venv_root), env_overrides=_UV_ENV_OVERRIDES)
        subprocess_post_check(process)
        return process

    def list_installed(
        self,
        *,
        venv_root: Path,
        venv_python: Path,
        not_required: bool = False,
    ) -> set[str]:
        if not_required:
            return list_not_required_packages(venv_python)

        cmd = self._uv_pip_command("list", venv_python, verbose=False)
        cmd += ["--format", "json"]
        process = run_subprocess(cmd, run_dir=str(venv_root), env_overrides=_UV_ENV_OVERRIDES)
        if process.returncode != 0:
            raise PipxError(
                f"Failed to execute {process.args}.\n"
                f"Process exited with return code {process.returncode}.\n"
                f"stderr: {process.stderr}"
            )
        return {entry["name"] for entry in json.loads(process.stdout.strip())}

    def run_raw_pip(
        self,
        *,
        venv_root: Path,
        venv_python: Path,
        args: list[str],
        capture_stdout: bool = True,
        capture_stderr: bool = True,
        verbose: bool = False,
    ) -> CompletedProcess[str]:
        # Mirror pip-backend ``pipx runpip TOOL`` (no args -> pip's help).
        cmd: list[str | Path] = [self._binary, "pip", args[0] if args else "--help"]
        cmd += ["--python", str(venv_python)]
        if verbose:
            cmd.append("--verbose")
        elif args:
            cmd.append("--quiet")
        cmd += _strip_pip_quiet_flags(args[1:])
        return run_subprocess(
            cmd,
            run_dir=str(venv_root),
            capture_stdout=capture_stdout,
            capture_stderr=capture_stderr,
            env_overrides=_UV_ENV_OVERRIDES,
        )

    def _uv_pip_command(self, subcommand: str, venv_python: Path, *, verbose: bool) -> list[str | Path]:
        cmd: list[str | Path] = [self._binary, "pip", subcommand, "--python", str(venv_python)]
        cmd.append("--verbose" if verbose else "--quiet")
        return cmd


def resolve_uv_binary() -> Path:
    # The version check fires here so non-Backend callers (e.g. ``pipx run``'s
    # uv-tool-run shortcut) still get the "needs uv >= X" message rather than
    # an opaque uv-side error.
    binary, _ = find_uv_binary()
    if binary is None:
        raise PipxError(
            "The uv backend was requested but the 'uv' executable could not be found.\n"
            "Install pipx with the uv extra (`pipx install pipx[uv]`) or place 'uv' on your PATH.\n"
            "Alternatively, run with `--backend pip` (or set PIPX_DEFAULT_BACKEND=pip)."
        )
    _check_uv_version(binary)
    return binary


@cache
def find_uv_binary() -> tuple[Path | None, str]:
    # Cached so ``upgrade-all`` and similar hot loops don't re-walk PATH per venv.
    if _FIND_UV_BIN_FROM_EXTRA is not None:
        try:
            bundled = Path(_FIND_UV_BIN_FROM_EXTRA())
        except FileNotFoundError:
            pass
        else:
            # ``is_file`` rejects a stale path from a half-installed extra;
            # ``_binary_runs`` catches the exec-fails case (missing dylib,
            # ENOEXEC) so we fall through to PATH instead of erroring later.
            if bundled.is_file() and _binary_runs(bundled):
                return bundled, "bundled"
    if path := shutil.which("uv"):
        return Path(path), "path"
    return None, "missing"


def _binary_runs(binary: Path) -> bool:
    # Liveness probe; full floor-version check stays in ``_check_uv_version``.
    try:
        process = subprocess.run([str(binary), "--version"], check=False, text=True, capture_output=True, timeout=10)
    except OSError as exc:
        _LOGGER.debug(f"uv launch probe failed for {binary}: {exc}")
        return False
    if process.returncode != 0 or not _VERSION_RE.search(process.stdout):
        _LOGGER.debug(
            f"uv launch probe rejected {binary}: "
            f"rc={process.returncode}, stdout={process.stdout!r}, stderr={process.stderr!r}"
        )
        return False
    return True


@cache
def _check_uv_version(binary: Path) -> Version:
    # Cached so ``upgrade-all`` over many venvs doesn't fork uv repeatedly.
    process = subprocess.run([str(binary), "--version"], check=False, text=True, capture_output=True)
    if (match := _VERSION_RE.search(process.stdout)) is None:
        raise PipxError(
            f"Could not parse uv version from {binary} "
            f"(rc={process.returncode}, stdout={process.stdout!r}, stderr={process.stderr!r})."
        )
    try:
        version = Version(match.group(1))
    except InvalidVersion as exc:
        raise PipxError(f"Unrecognized uv version {match.group(1)!r}.") from exc
    if version < _MIN_UV_VERSION:
        raise PipxError(
            f"pipx needs uv>={_MIN_UV_VERSION}, but {binary} reports {version}.\n"
            "Upgrade uv (`uv self update` or reinstall pipx[uv]), or run with `--backend pip` to bypass."
        )
    return version


def _strip_pip_quiet_flags(pip_args: list[str]) -> list[str]:
    return [arg for arg in pip_args if arg not in ("-q", "-qq", "--quiet")]


__all__ = [
    "UvBackend",
    "find_uv_binary",
    "resolve_uv_binary",
]



import logging
import subprocess
from pathlib import Path

from packaging import version

from pipx import commands, constants, paths, standalone_python
from pipx.animate import animate
from pipx.pipx_metadata_file import PipxMetadata
from pipx.util import is_paths_relative, rmdir
from pipx.venv import Venv, VenvContainer

logger = logging.getLogger(__name__)


def get_installed_standalone_interpreters() -> list[Path]:
    return [python_dir for python_dir in paths.ctx.standalone_python_cachedir.iterdir() if python_dir.is_dir()]


def get_venvs_using_standalone_interpreter(venv_container: VenvContainer) -> list[Venv]:
    venvs: list[Venv] = []
    for venv_dir in venv_container.iter_venv_dirs():
        venv = Venv(venv_dir)
        if venv.pipx_metadata.source_interpreter:
            venvs.append(venv)
    return venvs


def get_interpreter_users(interpreter: Path, venvs: list[Venv]) -> list[PipxMetadata]:
    return [
        venv.pipx_metadata
        for venv in venvs
        if venv.pipx_metadata.source_interpreter
        and is_paths_relative(venv.pipx_metadata.source_interpreter, interpreter)
    ]


def list_interpreters(
    venv_container: VenvContainer,
):
    interpreters = get_installed_standalone_interpreters()
    venvs = get_venvs_using_standalone_interpreter(venv_container)
    output: list[str] = []
    output.append(f"Standalone interpreters are in {paths.ctx.standalone_python_cachedir}")
    for interpreter in interpreters:
        output.append(f"Python {interpreter.name}")
        used_in = get_interpreter_users(interpreter, venvs)
        if used_in:
            output.append("    Used in:")
            output.extend(f"     - {p.main_package.package} {p.main_package.package_version}" for p in used_in)
        else:
            output.append("    Unused")

    print("\n".join(output))
    return constants.EXIT_CODE_OK


def prune_interpreters(
    venv_container: VenvContainer,
):
    interpreters = get_installed_standalone_interpreters()
    venvs = get_venvs_using_standalone_interpreter(venv_container)
    removed = []
    for interpreter in interpreters:
        if get_interpreter_users(interpreter, venvs):
            continue
        rmdir(interpreter, safe_rm=True)
        removed.append(interpreter.name)
    if removed:
        print("Successfully removed:")
        for interpreter_name in removed:
            print(f" - Python {interpreter_name}")
    else:
        print("Nothing to remove")
    return constants.EXIT_CODE_OK


def get_latest_micro_version(
    current_version: version.Version, latest_python_versions: list[version.Version]
) -> version.Version:
    for latest_python_version in latest_python_versions:
        if current_version.major == latest_python_version.major and current_version.minor == latest_python_version.minor:
            return latest_python_version
    return current_version


def upgrade_interpreters(venv_container: VenvContainer, verbose: bool):
    with animate("Getting the index of the latest standalone python builds", not verbose):
        latest_pythons = standalone_python.list_pythons(use_cache=False)

    parsed_latest_python_versions = []
    for latest_python_version in latest_pythons:
        try:
            parsed_latest_python_versions.append(version.parse(latest_python_version))
        except version.InvalidVersion:
            logger.info(f"Invalid version found in latest pythons: {latest_python_version}. Skipping.")

    upgraded = []

    for interpreter_dir in paths.ctx.standalone_python_cachedir.iterdir():
        if not interpreter_dir.is_dir():
            continue

        interpreter_python = interpreter_dir / "python.exe" if constants.WINDOWS else interpreter_dir / "bin" / "python3"
        interpreter_full_version = (
            subprocess.run([str(interpreter_python), "--version"], stdout=subprocess.PIPE, check=True, text=True)
            .stdout.strip()
            .split()[1]
        )
        try:
            parsed_interpreter_full_version = version.parse(interpreter_full_version)
        except version.InvalidVersion:
            logger.info(f"Invalid version found in interpreter at {interpreter_dir}. Skipping.")
            continue
        latest_micro_version = get_latest_micro_version(parsed_interpreter_full_version, parsed_latest_python_versions)
        if latest_micro_version > parsed_interpreter_full_version:
            standalone_python.download_python_build_standalone(
                f"{latest_micro_version.major}.{latest_micro_version.minor}",
                override=True,
            )

            for venv_dir in venv_container.iter_venv_dirs():
                venv = Venv(venv_dir)
                if venv.pipx_metadata.source_interpreter is not None and is_paths_relative(
                    venv.pipx_metadata.source_interpreter, interpreter_dir
                ):
                    print(
                        f"Upgrade the interpreter of {venv.name} from {interpreter_full_version} to {latest_micro_version}"
                    )
                    commands.reinstall(
                        venv_dir=venv_dir,
                        local_bin_dir=paths.ctx.bin_dir,
                        local_man_dir=paths.ctx.man_dir,
                        python=str(interpreter_python),
                        verbose=verbose,
                    )
                    upgraded.append((venv.name, interpreter_full_version, latest_micro_version))

    if upgraded:
        print("Successfully upgraded the interpreter(s):")
        for venv_name, old_version, new_version in upgraded:
            print(f" - {venv_name}: {old_version} -> {new_version}")
    else:
        print("Nothing to upgrade")

    # Any failure to upgrade will raise PipxError, otherwise success
    return constants.EXIT_CODE_OK



import os
import sys

if not __spec__ or not __spec__.parent:
    # Running from source. Add pipx's source code to the system
    # path to allow direct invocation, such as:
    #   python src/pipx --help
    pipx_package_source_path = os.path.dirname(os.path.dirname(__file__))
    sys.path.insert(0, pipx_package_source_path)

from pipx.main import cli

if __name__ == "__main__":
    sys.exit(cli())



import logging
import site
import sys
from pathlib import Path

import userpath  # type: ignore[import-not-found]

from pipx import paths
from pipx.constants import EXIT_CODE_OK, ExitCode
from pipx.emojis import hazard, stars
from pipx.util import pipx_wrap

logger = logging.getLogger(__name__)


def get_pipx_user_bin_path() -> Path | None:
    """Returns None if pipx is not installed using `pip --user`
    Otherwise returns parent dir of pipx binary
    """
    # NOTE: using this method to detect pip user-installed pipx will return
    #   None if pipx was installed as editable using `pip install --user -e`

    # https://docs.python.org/3/install/index.html#inst-alt-install-user
    #   Linux + Mac:
    #       scripts in <userbase>/bin
    #   Windows:
    #       scripts in <userbase>/Python<XY>/Scripts
    #       modules in <userbase>/Python<XY>/site-packages

    pipx_bin_path = None

    script_path = Path(__file__).resolve()
    userbase_path = Path(site.getuserbase()).resolve()
    try:
        _ = script_path.relative_to(userbase_path)
    except ValueError:
        pip_user_installed = False
    else:
        pip_user_installed = True
    if pip_user_installed:
        test_paths = (
            userbase_path / "bin" / "pipx",
            Path(site.getusersitepackages()).resolve().parent / "Scripts" / "pipx.exe",
        )
        for test_path in test_paths:
            if test_path.exists():
                pipx_bin_path = test_path.parent
                break

    return pipx_bin_path


def ensure_path(
    location: Path, *, force: bool, prepend: bool = False, all_shells: bool = False, dry_run: bool = False
) -> tuple[bool, bool]:
    """Ensure location is in user's PATH or add it to PATH.
    If prepend is True, location will be prepended to PATH, else appended.
    If dry_run is True, report what would change without modifying PATH or
    any shell configuration file.
    Returns True if location was added to PATH
    """
    location_str = str(location)
    path_added = False
    need_shell_restart = userpath.need_shell_restart(location_str)
    in_current_path = userpath.in_current_path(location_str)

    if force or (not in_current_path and not need_shell_restart):
        if dry_run:
            action = "prepend" if prepend else "append"
            print(
                pipx_wrap(
                    f"Would {action} {location_str} to the PATH environment variable.",
                    subsequent_indent=" " * 4,
                )
            )
            return (False, False)
        if prepend:
            path_added = userpath.prepend(location_str, "pipx", all_shells=all_shells)
        else:
            path_added = userpath.append(location_str, "pipx", all_shells=all_shells)
        if not path_added:
            print(
                pipx_wrap(
                    f"{hazard}  {location_str} is not added to the PATH environment variable successfully. "
                    f"You may need to add it to PATH manually.",
                    subsequent_indent=" " * 4,
                )
            )
        else:
            print(
                pipx_wrap(
                    f"Success! Added {location_str} to the PATH environment variable.",
                    subsequent_indent=" " * 4,
                )
            )
        need_shell_restart = userpath.need_shell_restart(location_str)
    elif not in_current_path and need_shell_restart:
        print(
            pipx_wrap(
                f"""
                {location_str} has been added to PATH, but you need to
                open a new terminal or re-login for this PATH change to take
                effect. Alternatively, you can source your shell's config file
                with e.g. 'source ~/.bashrc'.
                """,
                subsequent_indent=" " * 4,
            )
        )
    else:
        print(pipx_wrap(f"{location_str} is already in PATH.", subsequent_indent=" " * 4))

    return (path_added, need_shell_restart)


def ensure_pipx_paths(force: bool, prepend: bool = False, all_shells: bool = False, dry_run: bool = False) -> ExitCode:
    """Returns pipx exit code."""
    bin_paths = {paths.ctx.bin_dir}

    pipx_user_bin_path = get_pipx_user_bin_path()
    if pipx_user_bin_path is not None:
        bin_paths.add(pipx_user_bin_path)

    path_added = False
    need_shell_restart = False
    path_action_str = "prepended to" if prepend else "appended to"

    for bin_path in bin_paths:
        (path_added_current, need_shell_restart_current) = ensure_path(
            bin_path, prepend=prepend, force=force, all_shells=all_shells, dry_run=dry_run
        )
        path_added |= path_added_current
        need_shell_restart |= need_shell_restart_current

    print()

    if dry_run:
        print(pipx_wrap("This was a dry run; no changes were made to your PATH or shell configuration files."))
        return EXIT_CODE_OK

    if path_added:
        print(
            pipx_wrap(
                """
                Consider adding shell completions for pipx. Run 'pipx
                completions' for instructions.
                """
            )
            + "\n"
        )
    elif not need_shell_restart:
        sys.stdout.flush()
        logger.warning(
            pipx_wrap(
                f"""
                {hazard}  All pipx binary directories have been {path_action_str} PATH. If you
                are sure you want to proceed, try again with the '--force'
                flag.
                """
            )
            + "\n"
        )

    if need_shell_restart:
        print(
            pipx_wrap(
                """
                You will need to open a new terminal or re-login for the PATH
                changes to take effect. Alternatively, you can source your shell's
                config file with e.g. 'source ~/.bashrc'.
                """
            )
            + "\n"
        )

    print(f"Otherwise pipx is ready to go! {stars}")

    return EXIT_CODE_OK



import datetime
import logging
import os
import time
from configparser import ConfigParser
from contextlib import suppress
from pathlib import Path

from packaging.requirements import InvalidRequirement, Requirement
from packaging.specifiers import SpecifierSet

from pipx import paths
from pipx.animate import animate
from pipx.constants import WINDOWS
from pipx.emojis import strtobool
from pipx.interpreter import DEFAULT_PYTHON
from pipx.util import (
    get_site_packages,
    get_venv_paths,
    run_subprocess,
    subprocess_post_check,
)

logger = logging.getLogger(__name__)


SHARED_LIBS_MAX_AGE_SEC = datetime.timedelta(days=30).total_seconds()
DISABLE_SHARED_LIBS_AUTO_UPGRADE = "PIPX_DISABLE_SHARED_LIBS_AUTO_UPGRADE"


def shared_libs_auto_upgrade_disabled() -> bool:
    return strtobool(os.getenv(DISABLE_SHARED_LIBS_AUTO_UPGRADE, "0"))


def _venv_python_is_valid(python_path: Path) -> bool:
    """Check if a venv's Python is valid and its underlying interpreter exists.

    On Windows, a venv's python.exe is a wrapper that uses pyvenv.cfg to find
    the actual Python installation. If the original Python is uninstalled,
    the wrapper exists but cannot execute. This function checks that the
    underlying interpreter referenced in pyvenv.cfg still exists.
    """
    if not WINDOWS:
        return True

    pyvenv_cfg = python_path.parent.parent / "pyvenv.cfg"
    if not pyvenv_cfg.is_file():
        return True

    try:
        config = ConfigParser()
        with open(pyvenv_cfg, encoding="utf-8") as f:
            # ConfigParser needs a section header, pyvenv.cfg doesn't have one
            config.read_string("[DEFAULT]\n" + f.read())
        home = config.get("DEFAULT", "home", fallback=None)
        if home:
            # The home path points to the directory containing the original python.exe
            original_python = Path(home) / "python.exe"
            if not original_python.is_file():
                logger.info(f"Shared libs venv references a missing Python interpreter: {original_python}")
                return False
    except Exception:
        # If we can't read pyvenv.cfg, assume the venv is valid
        pass

    return True


class _SharedLibs:
    def __init__(self) -> None:
        self._site_packages: dict[Path, Path] = {}
        self.has_been_updated_this_run = False
        self.has_been_logged_this_run = False

    @property
    def root(self) -> Path:
        return paths.ctx.shared_libs

    @property
    def bin_path(self) -> Path:
        bin_path, _, _ = get_venv_paths(self.root)
        return bin_path

    @property
    def python_path(self) -> Path:
        _, python_path, _ = get_venv_paths(self.root)
        return python_path

    @property
    def man_path(self) -> Path:
        _, _, man_path = get_venv_paths(self.root)
        return man_path

    @property
    def pip_path(self) -> Path:
        return self.bin_path / ("pip" if not WINDOWS else "pip.exe")

    @property
    def site_packages(self) -> Path:
        if self.python_path not in self._site_packages:
            self._site_packages[self.python_path] = get_site_packages(self.python_path)

        return self._site_packages[self.python_path]

    def create(self, pip_args: list[str], verbose: bool = False) -> None:
        if not self.is_valid:
            with animate("creating shared libraries", not verbose):
                create_process = run_subprocess(
                    [DEFAULT_PYTHON, "-m", "venv", "--clear", self.root], run_dir=str(self.root)
                )
            subprocess_post_check(create_process)

            # ignore installed packages to ensure no unexpected patches from the OS vendor
            # are used
            pip_args = pip_args or []
            pip_args.append("--force-reinstall")
            self.upgrade(pip_args=pip_args, verbose=verbose, raises=True)

            # Remove setuptools from shared libs to prevent it from leaking into
            # app venvs via the .pth file. On Python < 3.12, venv bundles setuptools
            # via ensurepip, but a 3.10 setuptools breaks when imported under 3.12+
            # because distutils was removed from the stdlib.
            run_subprocess(
                [self.python_path, "-m", "pip", "--no-input", "uninstall", "-y", "setuptools"],
                capture_stderr=False,
            )

    @property
    def is_valid(self) -> bool:
        if self.python_path.is_file():
            # On Windows, check that the venv's underlying Python still exists
            if not _venv_python_is_valid(self.python_path):
                return False

            check_pip = "import importlib.util; print(importlib.util.find_spec('pip'))"
            out = run_subprocess(
                [self.python_path, "-c", check_pip],
                capture_stderr=False,
                log_cmd_str="<checking pip's availability>",
            ).stdout.strip()

            return self.pip_path.is_file() and out != "None"
        else:
            return False

    @property
    def needs_upgrade(self) -> bool:
        if self.has_been_updated_this_run:
            return False

        if not self.pip_path.is_file():
            return True

        now = time.time()
        time_since_last_update_sec = now - self.pip_path.stat().st_mtime
        if not self.has_been_logged_this_run:
            logger.info(
                f"Time since last upgrade of shared libs, in seconds: {time_since_last_update_sec:.0f}. "
                f"Upgrade will be run by pipx if greater than {SHARED_LIBS_MAX_AGE_SEC:.0f}."
            )
            self.has_been_logged_this_run = True
        return time_since_last_update_sec > SHARED_LIBS_MAX_AGE_SEC

    def upgrade(self, *, pip_args: list[str], verbose: bool = False, raises: bool = False) -> None:
        if not self.is_valid:
            self.create(verbose=verbose, pip_args=pip_args)
            return

        # Don't try to upgrade multiple times per run
        if self.has_been_updated_this_run:
            logger.info(f"Already upgraded libraries in {self.root}")
            return

        if pip_args is None:
            pip_args = []  # type: ignore[unreachable]

        logger.info(f"Upgrading shared libraries in {self.root}")

        ignored_args = ["--editable"]
        _pip_args = [arg for arg in pip_args if arg not in ignored_args]
        if not verbose:
            _pip_args.append("-q")

        user_pip_req = None
        for arg in _pip_args:
            with suppress(InvalidRequirement):
                if (req := Requirement(arg)).name == "pip":
                    user_pip_req = req
                    break

        add_default = not user_pip_req or not (user_pip_req.specifier & SpecifierSet(">=23.1"))
        install_args = [*_pip_args, "pip >= 23.1"] if add_default else _pip_args

        try:
            with animate("upgrading shared libraries", not verbose):
                upgrade_process = run_subprocess(
                    [
                        self.python_path,
                        "-m",
                        "pip",
                        "--no-input",
                        "--disable-pip-version-check",
                        "install",
                        "--upgrade",
                        *install_args,
                    ]
                )
            subprocess_post_check(upgrade_process)

            self.has_been_updated_this_run = True
            self.pip_path.touch()

        except Exception:
            logger.error("Failed to upgrade shared libraries", exc_info=not raises)
            if raises:
                raise


shared_libs = _SharedLibs()



import shutil
import textwrap


def pipx_wrap(text: str, subsequent_indent: str = "", keep_newlines: bool = False) -> str:
    """Dedent, strip, wrap to shell width. Don't break on hyphens, only spaces"""
    minimum_width = 40
    width = max(shutil.get_terminal_size((80, 40)).columns, minimum_width) - 2

    text = textwrap.dedent(text).strip()
    if keep_newlines:
        return "\n".join(
            [
                textwrap.fill(
                    line,
                    width=width,
                    subsequent_indent=subsequent_indent,
                    break_on_hyphens=False,
                )
                for line in text.splitlines()
            ]
        )
    else:
        return textwrap.fill(
            text,
            width=width,
            subsequent_indent=subsequent_indent,
            break_on_hyphens=False,
        )



from pathlib import Path

from pipx.constants import ExitCode
from pipx.util import PipxError
from pipx.venv import Venv


def run_pip(package: str, venv_dir: Path, pip_args: list[str], verbose: bool) -> ExitCode:
    """Returns pipx exit code."""
    venv = Venv(venv_dir, verbose=verbose)
    if not venv.python_path.exists():
        raise PipxError(f"venv for {package!r} was not found. Was {package!r} installed with pipx?")
    venv.verbose = True
    return venv.run_pip_get_exit_code(pip_args)



from __future__ import annotations

import json
import logging
from typing import TYPE_CHECKING, Final

from pipx.animate import animate
from pipx.backends._base import PIP, Backend
from pipx.constants import PIPX_SHARED_PTH
from pipx.shared_libs import shared_libs
from pipx.util import (
    PipxError,
    get_site_packages,
    get_venv_paths,
    run_subprocess,
    subprocess_post_check,
    subprocess_post_check_handle_pip_error,
)
from pipx.venv_inspect import list_not_required_packages

if TYPE_CHECKING:
    from pathlib import Path
    from subprocess import CompletedProcess

_LOGGER: Final[logging.Logger] = logging.getLogger(__name__)


class PipBackend(Backend):
    name = PIP

    def needs_shared_libs(self) -> bool:
        return True

    def upgrade_packaging_libraries(
        self,
        venv_python: Path,
        pip_args: list[str],
        *,
        verbose: bool,
    ) -> None:
        # Reached only for ``pipx install pip``-style venvs that ship pip
        # in-tree; shared-libs venvs upgrade through ``shared_libs.upgrade``.
        process = run_subprocess(
            [str(venv_python), "-m", "pip", "--no-input", "install", "--upgrade", *pip_args, "pip"],
            run_dir=str(venv_python.parent.parent),
        )
        subprocess_post_check(process)

    def create_venv(
        self,
        root: Path,
        *,
        python: str,
        venv_args: list[str],
        pip_args: list[str],
        include_pip: bool,
        verbose: bool,
    ) -> None:
        cmd = [python, "-m", "venv"]
        if not include_pip:
            cmd.append("--without-pip")
        cmd += [*venv_args, str(root)]
        with animate("creating virtual environment", not verbose):
            venv_process = run_subprocess(cmd, run_dir=str(root))
        subprocess_post_check(venv_process)

        shared_libs.create(verbose=verbose, pip_args=pip_args)
        if not include_pip:
            _, python_path, _ = get_venv_paths(root)
            pipx_pth = get_site_packages(python_path) / PIPX_SHARED_PTH
            pipx_pth.write_text(f"{shared_libs.site_packages}\n")

    def install(
        self,
        *,
        venv_root: Path,
        venv_python: Path,
        requirements: list[str],
        pip_args: list[str],
        no_deps: bool = False,
        upgrade: bool = False,
        log_pip_errors: bool = True,
        verbose: bool = False,
    ) -> CompletedProcess[str]:
        cmd: list[str] = [str(venv_python), "-m", "pip", "--no-input", "install"]
        if upgrade:
            cmd.append("--upgrade")
        if no_deps:
            cmd.append("--no-dependencies")
        cmd += [*pip_args, *requirements]
        process = run_subprocess(
            cmd,
            log_stdout=not log_pip_errors,
            log_stderr=not log_pip_errors,
            run_dir=str(venv_root),
        )
        if log_pip_errors:
            subprocess_post_check_handle_pip_error(process)
        return process

    def uninstall(
        self,
        *,
        venv_root: Path,
        venv_python: Path,
        package: str,
        verbose: bool,
    ) -> CompletedProcess[str]:
        cmd = [str(venv_python), "-m", "pip", "uninstall", "-y", package]
        if not verbose:
            cmd.append("-q")
        process = run_subprocess(cmd, run_dir=str(venv_root))
        subprocess_post_check(process)
        return process

    def list_installed(
        self,
        *,
        venv_root: Path,
        venv_python: Path,
        not_required: bool = False,
    ) -> set[str]:
        del venv_root
        if not_required:
            return list_not_required_packages(venv_python)
        cmd = [str(venv_python), "-m", "pip", "list", "--format=json"]
        process = run_subprocess(cmd)
        if process.returncode != 0:
            raise PipxError(
                f"Failed to execute {process.args}.\n"
                f"Process exited with return code {process.returncode}.\n"
                f"stderr: {process.stderr}"
            )
        return {entry["name"] for entry in json.loads(process.stdout.strip())}

    def run_raw_pip(
        self,
        *,
        venv_root: Path,
        venv_python: Path,
        args: list[str],
        capture_stdout: bool = True,
        capture_stderr: bool = True,
        verbose: bool = False,
    ) -> CompletedProcess[str]:
        cmd = [str(venv_python), "-m", "pip", *args]
        if not verbose:
            cmd.append("-q")
        return run_subprocess(
            cmd,
            capture_stdout=capture_stdout,
            capture_stderr=capture_stderr,
            run_dir=str(venv_root),
        )


__all__ = [
    "PipBackend",
]



import os

from pipx import paths
from pipx.backends import env_default_backend, find_uv_binary, resolve_backend_name
from pipx.constants import EXIT_CODE_OK, ExitCode
from pipx.emojis import EMOJI_SUPPORT
from pipx.interpreter import DEFAULT_PYTHON
from pipx.shared_libs import (
    DISABLE_SHARED_LIBS_AUTO_UPGRADE,
    shared_libs_auto_upgrade_disabled,
)
from pipx.util import PipxError

ENVIRONMENT_VARIABLES = [
    "PIPX_HOME",
    "PIPX_GLOBAL_HOME",
    "PIPX_BIN_DIR",
    "PIPX_GLOBAL_BIN_DIR",
    "PIPX_MAN_DIR",
    "PIPX_GLOBAL_MAN_DIR",
    "PIPX_SHARED_LIBS",
    "PIPX_DEFAULT_PYTHON",
    "PIPX_DEFAULT_BACKEND",
    "PIPX_FETCH_MISSING_PYTHON",
    "PIPX_FETCH_PYTHON",
    DISABLE_SHARED_LIBS_AUTO_UPGRADE,
    "PIPX_USE_EMOJI",
    "PIPX_HOME_ALLOW_SPACE",
]
DERIVED_ENVIRONMENT_VARIABLES = [
    "PIPX_LOCAL_VENVS",
    "PIPX_LOG_DIR",
    "PIPX_TRASH_DIR",
    "PIPX_VENV_CACHEDIR",
    "PIPX_STANDALONE_PYTHON_CACHEDIR",
    "PIPX_RESOLVED_BACKEND",
    "PIPX_BACKEND_SOURCE",
    "PIPX_UV_BINARY",
    "UV_CACHE_DIR",
]
ENVIRONMENT_VALUE_CHOICES = ENVIRONMENT_VARIABLES + DERIVED_ENVIRONMENT_VARIABLES


def environment(value: str) -> ExitCode:
    """Print a list of environment variables and paths used by pipx"""
    resolved_backend, backend_source = resolve_backend_name(env_value=env_default_backend())
    uv_binary, _uv_source = find_uv_binary()
    derived_values = {
        "PIPX_HOME": paths.ctx.home,
        "PIPX_BIN_DIR": paths.ctx.bin_dir,
        "PIPX_MAN_DIR": paths.ctx.man_dir,
        "PIPX_SHARED_LIBS": paths.ctx.shared_libs,
        "PIPX_LOCAL_VENVS": paths.ctx.venvs,
        "PIPX_LOG_DIR": paths.ctx.logs,
        "PIPX_TRASH_DIR": paths.ctx.trash,
        "PIPX_VENV_CACHEDIR": paths.ctx.venv_cache,
        "PIPX_STANDALONE_PYTHON_CACHEDIR": paths.ctx.standalone_python_cachedir,
        "PIPX_DEFAULT_PYTHON": DEFAULT_PYTHON,
        "PIPX_RESOLVED_BACKEND": resolved_backend,
        "PIPX_BACKEND_SOURCE": backend_source,
        "PIPX_UV_BINARY": str(uv_binary) if uv_binary else "",
        "UV_CACHE_DIR": os.environ.get("UV_CACHE_DIR", ""),
        DISABLE_SHARED_LIBS_AUTO_UPGRADE: str(shared_libs_auto_upgrade_disabled()).lower(),
        "PIPX_USE_EMOJI": str(EMOJI_SUPPORT).lower(),
        "PIPX_HOME_ALLOW_SPACE": str(paths.ctx.allow_spaces_in_home_path).lower(),
    }
    if value is None:
        print("Environment variables (set by user):")  # type: ignore[unreachable]
        print("")
        for env_variable in ENVIRONMENT_VARIABLES:
            env_value = os.getenv(env_variable, "")
            print(f"{env_variable}={env_value}")
        print("")
        print("Derived values (computed by pipx):")
        print("")
        for env_variable, derived_value in derived_values.items():
            print(f"{env_variable}={derived_value}")
    elif value in derived_values:
        print(derived_values[value])
    else:
        raise PipxError("Variable not found.")

    return EXIT_CODE_OK



import datetime
import hashlib
import json
import logging
import platform
import re
import shutil
import tarfile
import tempfile
import urllib.error
from functools import partial
from pathlib import Path
from typing import Any
from urllib.request import urlopen

from pipx import constants, paths
from pipx.animate import animate
from pipx.util import PipxError

logger = logging.getLogger(__name__)

# Much of the code in this module is adapted with extreme gratitude from
# https://github.com/tusharsadhwani/yen/blob/main/src/yen/github.py

MACHINE_SUFFIX: dict[str, dict[str, Any]] = {
    "Darwin": {
        "arm64": ["aarch64-apple-darwin-install_only.tar.gz"],
        "x86_64": ["x86_64-apple-darwin-install_only.tar.gz"],
    },
    "Linux": {
        "aarch64": {
            "glibc": ["aarch64-unknown-linux-gnu-install_only.tar.gz"],
            # musl doesn't exist
        },
        "x86_64": {
            "glibc": [
                "x86_64_v3-unknown-linux-gnu-install_only.tar.gz",
                "x86_64-unknown-linux-gnu-install_only.tar.gz",
            ],
            "musl": ["x86_64_v3-unknown-linux-musl-install_only.tar.gz"],
        },
    },
    "Windows": {"AMD64": ["x86_64-pc-windows-msvc-install_only.tar.gz"]},
}

GITHUB_API_URL = "https://api.github.com/repos/astral-sh/python-build-standalone/releases/latest"
PYTHON_VERSION_REGEX = re.compile(r"cpython-(\d+\.\d+\.\d+)")


def download_python_build_standalone(python_version: str, override: bool = False):
    """When all other python executable resolutions have failed,
    attempt to download and use an appropriate python build
    from https://github.com/astral-sh/python-build-standalone
    and unpack it into the pipx shared directory."""

    # python_version can be a bare version number like "3.10" or a "binary name" like python3.10
    # we'll convert it to a bare version number
    python_version = re.sub(r"[c]?python", "", python_version)

    install_dir = paths.ctx.standalone_python_cachedir / python_version
    installed_python = install_dir / "python.exe" if constants.WINDOWS else install_dir / "bin" / "python3"

    if override:
        if install_dir.exists():
            shutil.rmtree(install_dir)
    else:
        if installed_python.exists():
            return str(installed_python)

        if install_dir.exists():
            logger.warning(f"A previous attempt to install python {python_version} failed. Retrying.")
            shutil.rmtree(install_dir)

    full_version, (download_link, digest) = resolve_python_version(python_version)

    with tempfile.TemporaryDirectory() as tempdir:
        archive = Path(tempdir) / f"python-{full_version}.tar.gz"
        download_dir = Path(tempdir) / "download"

        # download the python build gz
        _download(full_version, download_link, archive)

        # unpack the python build
        _unpack(full_version, download_link, archive, download_dir, digest)

        # the python installation we want is nested in the tarball
        # under a directory named 'python'. We move it to the install
        # directory
        extracted_dir = download_dir / "python"
        shutil.move(extracted_dir, install_dir)

    return str(installed_python)


def _download(full_version: str, download_link: str, archive: Path):
    with animate(f"Downloading python {full_version} build", True):
        try:
            # python standalone builds are typically ~32MB in size. to avoid
            # ballooning memory usage, we read the file in chunks
            with urlopen(download_link) as response, open(archive, "wb") as file_handle:
                for data in iter(partial(response.read, 32768), b""):
                    file_handle.write(data)
        except urllib.error.URLError as e:
            raise PipxError(f"Unable to download python {full_version} build.") from e


def _unpack(full_version, download_link, archive: Path, download_dir: Path, expected_checksum: str):
    with animate(f"Unpacking python {full_version} build", True):
        # Calculate checksum efficiently
        sha256_hash = hashlib.sha256()
        with open(archive, "rb") as python_zip:
            # Read in chunks to avoid loading the whole file into memory
            for chunk in iter(lambda: python_zip.read(32768), b""):
                sha256_hash.update(chunk)

        checksum = "sha256:" + sha256_hash.hexdigest()

        # Validate checksum
        if checksum != expected_checksum:
            raise PipxError(
                f"Checksum mismatch for python {full_version} build. Expected {expected_checksum}, got {checksum}."
            )

        with tarfile.open(archive, mode="r:gz") as tar:
            tar.extractall(download_dir)


def _is_valid_python_index(index: Any) -> bool:
    if not isinstance(index, dict):
        return False

    fetched = index.get("fetched")
    releases = index.get("releases")
    if not isinstance(fetched, int | float) or not isinstance(releases, list):
        return False

    return all(
        isinstance(release, (list, tuple)) and len(release) == 2 and all(isinstance(value, str) for value in release)
        for release in releases
    )


def get_or_update_index(use_cache: bool = True):
    """Get or update the index of available python builds from
    the python-build-standalone repository."""
    index_file = paths.ctx.standalone_python_cachedir / "index.json"
    if use_cache and index_file.exists():
        index = json.loads(index_file.read_text())
        # Refresh legacy URL-only indexes, and update current indexes after 30 days.
        if _is_valid_python_index(index):
            fetched = datetime.datetime.fromtimestamp(index["fetched"])
            if datetime.datetime.now() - fetched > datetime.timedelta(days=30):
                index = {}
        else:
            index = {}
    else:
        index = {}
    if not index:
        releases = get_latest_python_releases()
        index = {"fetched": datetime.datetime.now().timestamp(), "releases": releases}
        # update index
        index_file.write_text(json.dumps(index))
    return index


def get_latest_python_releases() -> list[tuple[str, str]]:
    """Returns the list of python download links from the latest github release."""
    try:
        with urlopen(GITHUB_API_URL) as response:
            release_data = json.load(response)

    except urllib.error.URLError as e:
        # raise
        raise PipxError(f"Unable to fetch python-build-standalone release data (from {GITHUB_API_URL}).") from e

    return [(asset["browser_download_url"], asset["digest"]) for asset in release_data["assets"]]


def list_pythons(use_cache: bool = True) -> dict[str, tuple[str, str]]:
    """Returns available python versions for your machine and their download links."""
    system, machine = platform.system(), platform.machine()
    download_link_suffixes = MACHINE_SUFFIX[system][machine]
    # linux suffixes are nested under glibc or musl builds
    if system == "Linux":
        # fallback to musl if libc version is not found
        libc_version = platform.libc_ver()[0] or "musl"
        download_link_suffixes = download_link_suffixes[libc_version]

    python_releases = get_or_update_index(use_cache)["releases"]

    available_python_links = [
        (link, digest)
        # Suffixes are in order of preference.
        for download_link_suffix in download_link_suffixes
        for link, digest in python_releases
        if link.endswith(download_link_suffix)
    ]

    python_versions: dict[str, tuple[str, str]] = {}
    for link, digest in available_python_links:
        match = PYTHON_VERSION_REGEX.search(link)
        assert match is not None
        python_version = match[1]
        # Don't override already found versions, they are in order of preference
        if python_version in python_versions:
            continue

        python_versions[python_version] = link, digest

    return {
        version: python_versions[version]
        for version in sorted(
            python_versions,
            # sort by semver
            key=lambda version: [int(k) for k in version.split(".")],
            reverse=True,
        )
    }


def resolve_python_version(requested_version: str):
    pythons = list_pythons()
    requested_release = requested_version.split(".")

    for full_version, download_link in pythons.items():
        standalone_release = full_version.split(".")
        if requested_release == standalone_release[: len(requested_release)]:
            return full_version, download_link

    raise PipxError(f"Unable to acquire a standalone python build matching {requested_version}.")



import logging
import os
from importlib import metadata
from pathlib import Path

from packaging.utils import canonicalize_name

from pipx.colors import bold
from pipx.commands.uninstall import (
    _get_package_bin_dir_app_paths,
    _get_package_man_paths,
)
from pipx.constants import (
    EXIT_CODE_OK,
    EXIT_CODE_UNINJECT_ERROR,
    MAN_SECTIONS,
    ExitCode,
)
from pipx.emojis import stars
from pipx.util import PipxError, pipx_wrap
from pipx.venv import Venv
from pipx.venv_inspect import fetch_info_in_venv, get_dist, get_required_dependency_names

logger = logging.getLogger(__name__)


def uninject(
    venv_dir: Path,
    dependencies: list[str],
    *,
    local_bin_dir: Path,
    local_man_dir: Path,
    leave_deps: bool,
    verbose: bool,
) -> ExitCode:
    """Returns pipx exit code"""

    if not venv_dir.exists() or next(venv_dir.iterdir(), None) is None:
        raise PipxError(f"Virtual environment {venv_dir.name} does not exist.")

    venv = Venv(venv_dir, verbose=verbose)

    if not venv.package_metadata:
        raise PipxError(
            f"""
            Can't uninject from Virtual Environment {venv_dir.name!r}.
            {venv_dir.name!r} has missing internal pipx metadata.
            It was likely installed using a pipx version before 0.15.0.0.
            Please uninstall and install {venv_dir.name!r} manually to fix.
            """
        )

    all_success = True
    for dep in dependencies:
        all_success &= uninject_dep(
            venv,
            dep,
            local_bin_dir=local_bin_dir,
            local_man_dir=local_man_dir,
            leave_deps=leave_deps,
        )

    if all_success:
        return EXIT_CODE_OK
    else:
        return EXIT_CODE_UNINJECT_ERROR


def uninject_dep(
    venv: Venv,
    package_name: str,
    *,
    local_bin_dir: Path,
    local_man_dir: Path,
    leave_deps: bool = False,
) -> bool:
    package_name = canonicalize_name(package_name)

    if package_name == venv.pipx_metadata.main_package.package:
        logger.warning(
            pipx_wrap(
                f"""
            {package_name} is the main package of {venv.root.name}
            venv. Use `pipx uninstall {venv.root.name}` to uninstall instead of uninject.
            """,
                subsequent_indent=" " * 4,
            )
        )
        return False

    if package_name not in venv.pipx_metadata.injected_packages:
        logger.warning(f"{package_name} is not in the {venv.root.name} venv. Skipping.")
        return False

    need_app_uninstall = venv.package_metadata[package_name].include_apps

    new_resource_paths = get_include_resource_paths(package_name, venv, local_bin_dir, local_man_dir)

    if not leave_deps:
        orig_not_required_packages = venv.list_installed_packages(not_required=True)
        logger.info(f"Original not required packages: {orig_not_required_packages}")

    venv.uninstall_package(package=package_name, was_injected=True)

    if not leave_deps:
        new_not_required_packages = venv.list_installed_packages(not_required=True)
        logger.info(f"New not required packages: {new_not_required_packages}")

        deps_of_uninstalled = new_not_required_packages - orig_not_required_packages
        if deps_of_uninstalled:
            remaining_deps = _get_remaining_dependencies(venv, package_name)
            deps_of_uninstalled -= remaining_deps
            logger.info(f"Dependencies of uninstalled package: {deps_of_uninstalled}")

        for dep_package_name in deps_of_uninstalled:
            venv.uninstall_package(package=dep_package_name, was_injected=False)

        deps_string = " and its dependencies"
    else:
        deps_string = ""

    if need_app_uninstall:
        for path in new_resource_paths:
            try:
                os.unlink(path)
            except FileNotFoundError:
                logger.info(f"tried to remove but couldn't find {path}")
            else:
                logger.info(f"removed file {path}")

    print(f"Uninjected package {bold(package_name)}{deps_string} from venv {bold(venv.root.name)} {stars}")
    return True


def get_include_resource_paths(package_name: str, venv: Venv, local_bin_dir: Path, local_man_dir: Path) -> set[Path]:
    bin_dir_app_paths = _get_package_bin_dir_app_paths(
        venv, venv.package_metadata[package_name], venv.bin_path, local_bin_dir
    )
    man_paths = set()
    for man_section in MAN_SECTIONS:
        man_paths |= _get_package_man_paths(
            venv,
            venv.package_metadata[package_name],
            venv.man_path / man_section,
            local_man_dir / man_section,
        )

    pkg_metadata = venv.package_metadata[package_name]
    all_apps = set(pkg_metadata.apps)
    all_man_pages = set(pkg_metadata.man_pages)
    if pkg_metadata.include_dependencies:
        all_apps.update(pkg_metadata.apps_of_dependencies)
        all_man_pages.update(pkg_metadata.man_pages_of_dependencies)

    need_to_remove = set()
    for bin_dir_app_path in bin_dir_app_paths:
        if bin_dir_app_path.name in all_apps:
            need_to_remove.add(bin_dir_app_path)
    for man_path in man_paths:
        path = Path(man_path.parent.name) / man_path.name
        if str(path) in all_man_pages:
            need_to_remove.add(path)

    return need_to_remove


def _get_remaining_dependencies(venv: Venv, excluded_package: str) -> set[str]:
    venv_sys_path, venv_env, _ = fetch_info_in_venv(venv.python_path)
    distributions = tuple(metadata.distributions(path=venv_sys_path))

    remaining_deps: set[str] = set()
    remaining_packages: list[str] = [
        name
        for name in [venv.pipx_metadata.main_package.package] + list(venv.pipx_metadata.injected_packages)
        if name is not None and name != excluded_package
    ]
    for pkg_name in remaining_packages:
        remaining_deps |= _collect_transitive_deps(pkg_name, distributions, venv_env)

    return remaining_deps


def _collect_transitive_deps(
    package_name: str,
    distributions: tuple[metadata.Distribution, ...],
    env: dict[str, str],
    visited: set[str] | None = None,
) -> set[str]:
    if visited is None:
        visited = set()
    canonical = canonicalize_name(package_name)
    if canonical in visited:
        return visited
    visited.add(canonical)
    if dist := get_dist(package_name, distributions):
        for dep_name in get_required_dependency_names(dist, env):
            _collect_transitive_deps(dep_name, distributions, env, visited)
    return visited
